from collections import defaultdict
from datetime import datetime
import json
import os
import pickle
import posixpath
import re
import requests
import sys

import dateutil.parser
import fastremap
import numpy as np

from .. import compression
from .. import exceptions
from ..cacheservice import CacheService
from ..lib import Bbox, Vec, toiter
from ..storage import SimpleStorage, Storage, reset_connection_pools
from ..volumecutout import VolumeCutout
from ..datasource.graphene.metadata import GrapheneApiVersion

from .precomputed import CloudVolumePrecomputed

def warn(text):
  print(colorize('yellow', text))

def to_unix_time(timestamp):
  """
  Accepts integer UNIX timestamps, ISO 8601 datetime strings,
  and Python datetime objects and returns them as the equivalent
  UNIX timestamp or None if timestamp is None.
  """
  if isinstance(timestamp, str):
    timestamp = dateutil.parser.parse(timestamp) # returns datetime
  if isinstance(timestamp, datetime): # NB. do not change to elif
    timestamp = datetime.timestamp(timestamp)

  if not isinstance(timestamp, int) and timestamp is not None:
    raise ValueError("Not able to convert {} to UNIX time.".format(timestamp))

  return timestamp

class CloudVolumeGraphene(CloudVolumePrecomputed):

  @property
  def manifest_endpoint(self):
    return self.meta.manifest_endpoint

  @property
  def graph_chunk_size(self):
    return self.meta.graph_chunk_size 

  @property
  def mesh_chunk_size(self):
    # TODO: add this as new parameter to the info as it can be different from the chunkedgraph chunksize
    return self.meta.mesh_chunk_size
  
  def download_point(
    self, pt, size=256, 
    mip=None, parallel=None, 
    coord_resolution=None,
    **kwargs
  ):
    """
    Download to the right of point given in mip 0 coords.
    Useful for quickly visualizing a neuroglancer coordinate
    at an arbitary mip level.

    pt: (x,y,z)
    size: int or (sx,sy,sz)
    mip: int representing resolution level
    parallel: number of processes to launch (0 means all cores)
    coord_resolution: (rx,ry,rz) the coordinate resolution of the input point.
      Sometimes Neuroglancer is working in the resolution of another
      higher res layer and this can help correct that.

    Also accepts the arguments for download such as segids and preserve_zeros.

    Return: image as VolumeCutout(ndarray)
    """
    if isinstance(size, int) or isinstance(size, float):
      size = Vec(size, size, size)
    else:
      size = Vec(*size)

    if mip is None:
      mip = self.mip

    mip = self.meta.to_mip(mip)
    size2 = size // 2

    if coord_resolution is not None:
      factor = self.meta.resolution(0) / Vec(*coord_resolution)
      pt = Vec(*pt) / factor

    pt = self.point_to_mip(pt, mip=0, to_mip=mip)

    if all(size == 1):
      bbox = Bbox(pt, pt + 1).astype(np.int64)
    else:
      bbox = Bbox(pt - size2, pt + size2).astype(np.int64)

    if parallel is None:
      parallel = self.parallel

    return self.download(bbox, mip, parallel=parallel, **kwargs)

  def download(
    self, bbox, mip=None, 
    parallel=None, segids=None,
    preserve_zeros=False,
    agglomerate=False, timestamp=None,
    stop_layer=None
  ):
    """
    Downloads base segmentation and optionally agglomerates
    labels based on information in the graph server.

    bbox: specifies cutout to fetch
    mip: which resolution level to get (default self.mip)
    parallel: what parallel level to use (default self.parallel)

    agglomerate: if true, remap all watershed ids in the volume
      and return a flat segmentation.

    if agglomerate is true these options are available:

    timestamp: (agglomerate only) get the roots from this date and time
      formats accepted:
        int: unix timestamp
        datetime: self explainatory
        string: ISO 8601 date
    stop_layer: (agglomerate only) (int) if specified, return the lowest 
      parent at or above that layer. If not specified, go all the way 
      to the root id. 
        Layer 1: Watershed
        Layer 2: Within-Chunk Agglomeration
        Layer 2+: Between chunk interconnections (skip connections possible)

    if agglomerate is false, these other options come into play:

    segids: agglomerate the leaves of these segids from the graph 
      server and label them with the given segid.
    preserve_zeros: If segids is not None:
      False: mask other segids with zero
      True: mask other segids with the largest integer value
        contained by the image data type and leave zero as is.

    Returns: img as a VolumeCutout
    """
    if type(bbox) is Vec:
      bbox = Bbox(bbox, bbox+1)
    
    bbox = Bbox.create(
      bbox, context=self.bounds, 
      bounded=self.bounded, 
      autocrop=self.autocrop
    )
  
    if bbox.subvoxel():
      raise exceptions.EmptyRequestException("Requested {} is smaller than a voxel.".format(bbox))

    if (agglomerate and stop_layer is not None) and (stop_layer <= 0 or stop_layer > self.meta.n_layers):
      raise ValueError("Stop layer {} must be 1 <= stop_layer <= {} or None.".format(stop_layer, self.meta.n_layers))

    if mip is None:
      mip = self.mip

    mip0_bbox = self.bbox_to_mip(bbox, mip=mip, to_mip=0)
    # Only ever necessary to make requests within the bounding box
    # to the server. We can fill black in other situations.
    mip0_bbox = bbox.intersection(self.meta.bounds(0), mip0_bbox)

    img = super(CloudVolumeGraphene, self).download(bbox, mip=mip, parallel=parallel)

    if agglomerate:
      img = self.agglomerate_cutout(img, timestamp=timestamp, stop_layer=stop_layer)
      return VolumeCutout.from_volume(self.meta, mip, img, bbox)

    if segids is None:
      return img

    segids = list(toiter(segids))

    remapping = {}
    for segid in segids:
      leaves = self.get_leaves(segid, mip0_bbox, 0)
      remapping.update({ leaf: segid for leaf in leaves })
    
    img = fastremap.remap(img, remapping, preserve_missing_labels=True, in_place=True)

    mask_value = 0
    if preserve_zeros:
      mask_value = np.inf
      if np.issubdtype(self.dtype, np.integer):
        mask_value = np.iinfo(self.dtype).max

      segids.append(0)

    img = fastremap.mask_except(img, segids, in_place=True, value=mask_value)

    return VolumeCutout.from_volume(
      self.meta, mip, img, bbox 
    )
  
  def agglomerate_cutout(self, img, timestamp=None, stop_layer=None):
    """Remap a graphene volume to its latest root ids. This creates a flat segmentation."""
    if np.all(img == self.image.background_color) or stop_layer == 1:
      return img

    labels = fastremap.unique(img)
    if labels.size and labels[0] == 0:
      labels = labels[1:]

    roots = self.get_roots(labels, timestamp=timestamp, binary=True, stop_layer=stop_layer)
    mapping = { segid: root for segid, root in zip(labels, roots) }
    return fastremap.remap(img, mapping, preserve_missing_labels=True, in_place=True)

  def __getitem__(self, slices):
    return self.download(
      slices, mip=self.mip,
      preserve_zeros=True,
      parallel=self.parallel, 
    )

  def get_chunk_layer(self, node_or_chunk_id):
    """
    Extract Layer from Node ID or Chunk ID
    
    Returns: (int) layer number
    """
    return int(self.meta.decode_layer_id(node_or_chunk_id))

  def get_root(self, segid, *args, **kwargs):
    """Deprecated. Get a single root id for a single segid."""
    return get_roots(segid, *args, **kwargs)[0]

  def get_roots(self, segids, timestamp=None, binary=True, stop_layer=None):
    """
    Get the root ids for these labels.

    segids: (int or iterable) one or more segids to remap
    timestamp: get the roots from this date and time
      formats accepted:
        int: unix timestamp
        datetime: self explainatory
        string: ISO 8601 date
    binary: if true, send and receive segids as a 
      binary stream else, use JSON. The difference can
      be a 2x difference in bandwidth used.
    stop_layer: (int) if specified, return the lowest parent at or above 
      that layer. If not specified, go all the way to the root id. 
        Layer 1: Watershed
        Layer 2: Within-Chunk Agglomeration
        Layer 2+: Between chunk interconnections (skip connections possible)
    """
    segids = toiter(segids)
    if isinstance(segids, np.ndarray):
      segids = segids.tolist()

    try:
      segids.remove(0) # background segid
    except ValueError:
      pass

    timestamp = to_unix_time(timestamp)

    if stop_layer is not None:
      stop_layer = int(stop_layer)
      if stop_layer < 1 or stop_layer > self.meta.n_layers:
        raise ValueError("stop_layer ({}) must be between 1 and {} inclusive.".format(
          stop_layer, self.meta.n_layers
        ))

    if self.meta.supports_api('v1'):
      roots = self._get_roots_v1(segids, timestamp, binary, stop_layer)
    elif self.meta.supports_api('1.0'):
      roots = self._get_roots_legacy(segids, timestamp)
    else:
      raise exceptions.UnsupportedGrapheneAPIVersionError(
        "{} is not a supported API version. Supported versions: ".format(self.meta.api_version) \
        + ", ".join([ str(_) for _ in self.meta.supported_api_versions ])
      )

    return np.array(roots, dtype=self.meta.dtype)

  def get_chunk_mappings(self, chunk_id, timestamp=None):
    """
    Get the mapping of segments in a chunk at a given chunk graph layer 
    to their L1 watershed components.

    NOTE: Only L2 chunks are supported at this time.

    Required:
      chunk_id: uint64 chunk id (ie. an graphene label with a zeroed segid component)
        NOTE: This function actually accepts any graphene label and automatically converts
        it to a chunk ID before querying the graph server by zeroing out its segid component.
    Optional:
      timestamp: query the state of the graph server at the time point specified
        by a UNIX timestamp, ISO 8601 datetime string, or a python datetime object.

    Returns: {  
      chunk_label: [ watershed labels ],
      ... e.g. ...
      173729460028178433: [79450023430979610, 79450023431072298, ... ]
    }
    """
    timestamp = to_unix_time(timestamp)

    if not self.meta.supports_api('v1'):
      raise exceptions.UnsupportedGrapheneAPIVersionError(
        "{} is not a supported API version for range read requests. Currently, only version 1.0 is supported: ".format(self.meta.api_version) \
      )

    layer_id = self.meta.decode_layer_id(chunk_id)
    if layer_id != 2:
      raise ValueError("This function currently only accepts Layer 2 chunk IDs. Got {}".format(self.meta.decode_label(chunk_id)))

    chunk_id = self.meta.decode_chunk_id(chunk_id)
    
    version = GrapheneApiVersion('v1')
    path = version.path(self.meta.server_path)
    url = posixpath.join(self.meta.base_path, path, "l2_chunk_children_binary", str(chunk_id))

    params = {'as_array': True}
    if timestamp is not None:
      params['timestamp'] = timestamp

    response = requests.get(url, params=params, headers=self.meta.auth_header)
    response.raise_for_status()

    chunk_array = np.frombuffer(response.content, dtype=np.uint64)
    chunk_mappings = defaultdict(list)

    for i in range(0, len(chunk_array), 2):
      chunk_mappings[chunk_array[i]].append(chunk_array[i+1])

    return chunk_mappings

  def _get_roots_v1(self, segids, timestamp, binary=False, stop_layer=None):
    args = {}
    if timestamp is not None:
      args['timestamp'] = timestamp

    headers = {}
    headers.update(self.meta.auth_header)

    gzip_condition = len(segids) * 8 > 1e6

    if gzip_condition:
      headers['Content-Encoding'] = 'gzip'
      headers['Accept-Encoding'] = 'gzip;q=1, identity;q=0.1'
    else:
      headers['Accept-Encoding'] = 'identity'

    version = GrapheneApiVersion('v1')
    path = version.path(self.meta.server_path)

    params = {}
    if stop_layer:
      params['stop_layer'] = int(stop_layer)

    if binary:
      url = posixpath.join(self.meta.base_path, path, "roots_binary")
      data = np.array(segids, dtype=np.uint64).tobytes()
      params['timestamp'] = timestamp
    else:
      url = posixpath.join(self.meta.base_path, path, "roots")
      args['node_ids'] = segids
      data = json.dumps(args).encode('utf8')

    if gzip_condition:
      data = compression.compress(data, method='gzip')
    
    response = requests.post(url, data=data, headers=headers, params=params)
    response.raise_for_status()

    if binary:
      return np.frombuffer(response.content, dtype=np.uint64)
    else:
      return json.loads(response.content)['root_ids']

  def _get_roots_legacy(self, segids, timestamp):
    args = {}
    if timestamp is not None:
      args['timestamp'] = timestamp

    version = GrapheneApiVersion('1.0')
    path = version.path(self.meta.server_path)
    roots = []
    for segid in segids:
      url = posixpath.join(self.meta.base_path, path, "graph/{:d}/root".format(int(segid)))
      response = requests.get(url, json=args, headers=self.meta.auth_header)
      response.raise_for_status()
      root = np.frombuffer(response.content, dtype=np.uint64)[0]
      roots.append(root)
    return roots

  def get_leaves(self, root_id, bbox, mip):
    """
    get the supervoxels for this root_id

    params
    ------
    root_id: uint64 root id to find supervoxels for
    bbox: cloudvolume.lib.Bbox 3d bounding box for segmentation
    """
    root_id = int(root_id)
    url = posixpath.join(self.meta.server_url, "segment", str(root_id), "leaves")
    bbox = Bbox.create(bbox, context=self.meta.bounds(mip), bounded=self.bounded)
    response = requests.post(url, json=[ root_id ], params={
      'bounds': bbox.to_filename(),
    }, headers=self.meta.auth_header)
    response.raise_for_status()

    return np.frombuffer(response.content, dtype=np.uint64)



from collections import defaultdict
from copy import deepcopy
import time
import re
import json
import socket
from operator import mul
from functools import reduce

import numpy as np
from cloudvolume import CloudVolume

N = 3

CHUNK_SIZES = ( 
	(128,128,1),
	(128,128,2),
	(128,128,4),
	(128,128,8),
	(128,128,16),
	(128,128,32),
	(128,128,64),
	(128,128,128),
	(1024,1024,1),
	(1024,1024,2),
	(1024,1024,4),
	(1024,1024,8),
	(1024,1024,16),
	(1024,1024,32),
	(1024,1024,64),
	(1024,1024,100), # max size of SNEMI3D
)

logfile = open('./benchmark.tsv', 'wt')
logfile.write("hostname\tdirection\tcompression\timage_type\tchunk_size\tdtype\tMB\tMean MB/sec\tN\tmean (sec)\tfastest (sec)\tslowest (sec)\tcloudpath\n")
logfile.flush()

def stopwatch(fn):
	start = time.time()
	fn()
	return time.time() - start

def benchmark(fn, N):
	lst = [ stopwatch(fn) for _ in range(N) ]
	return {
		'mean': (sum(lst) / N), 
		'fastest': min(lst), 
		'slowest': max(lst), 
		'N': N,
	}

def MBs(vol, sec):
	dtype_bytes = np.dtype(vol.dtype).itemsize
	total_bytes = vol.bounds.volume() * dtype_bytes
	return total_bytes / 1e6 / sec

def disp(desc, vol, stats):
	mbs = lambda sec: MBs(vol, sec)
	return "%s -- mean=%.2f MB/sec, range=%.2f to %.2f MB/sec; N=%d" % (
		desc, mbs(stats['mean']), mbs(stats['slowest']), mbs(stats['fastest']), stats['N']
	)

def log(row):
	global logfile
	dtype_bytes = np.dtype(row['dtype']).itemsize
	row['MB'] = reduce(mul, row['chunk_size']) * dtype_bytes / 1024**2
	row['MBs'] = row['MB'] / row['mean']

	for k,v in row.items():
		if type(v) is float:
			row[k] = '%.3f' % v

	entry = "{hostname}\t{direction}\t{compression}\t{image_type}\t{chunk_size}\t{dtype}\t{MB}\t{MBs}\t{N}\t{mean}\t{fastest}\t{slowest}\t{cloudpath}\n".format(**row)
	logfile.write(entry)
	logfile.flush()

def benchmark_upload(voltype):
	global CHUNK_SIZES
	global N

	originvol = CloudVolume('gs://seunglab-test/test_v0/{}'.format(voltype))
	originvol.reset_scales()
	info = deepcopy(originvol.info)

	img = originvol[:]

	for chunk_size in CHUNK_SIZES[::-1]:
		cloudpath = 'gs://seunglab-test/test_v0/{}_upload_{}_{}_{}'.format(voltype, *chunk_size)
		info['scales'][0]['chunk_sizes'] = [ list(chunk_size) ]
		vol = CloudVolume(cloudpath, progress=True, info=info, compress='gzip')

		def upload():
			vol[:] = img

		stats = benchmark(upload, N)

		log({
			"direction": "upload",
			"compression": "gzip",
			"image_type": voltype,
			"N": N,
			"mean": stats['mean'],
			"fastest": stats['fastest'],
			"slowest": stats['slowest'],
			"cloudpath": cloudpath,
			"chunk_size": chunk_size,
			"dtype": vol.dtype,
			"hostname": socket.gethostname(),
		})

		vol.delete(vol.bounds)

	for chunk_size in CHUNK_SIZES[::-1]:
		cloudpath = 'gs://seunglab-test/test_v0/{}_upload_{}_{}_{}'.format(voltype, *chunk_size)
		info['scales'][0]['chunk_sizes'] = [ list(chunk_size) ]
		vol = CloudVolume(cloudpath, progress=True, info=info, compress='')

		def upload():
			vol[:] = img
		
		stats = benchmark(upload, N)

		log({
			"direction": "upload",
			"compression": "none",
			"image_type": voltype,
			"N": N,
			"mean": stats['mean'],
			"fastest": stats['fastest'],
			"slowest": stats['slowest'],
			"cloudpath": cloudpath,
			"chunk_size": chunk_size,
			"dtype": vol.dtype,
			"hostname": socket.gethostname(),
		})

		vol.delete(vol.bounds)


def benchmark_download(voltype):
	global CHUNK_SIZES
	global N

	for chunk_size in CHUNK_SIZES[::-1]:
		cloudpath = 'gs://seunglab-test/test_v0/{}_{}_{}_{}'.format(voltype, *chunk_size)
		vol = CloudVolume(cloudpath, progress=True)
		stats = benchmark(lambda: vol[:], N)

		log({
			"direction": "download",
			"compression": "gzip",
			"image_type": voltype,
			"N": N,
			"mean": stats['mean'],
			"fastest": stats['fastest'],
			"slowest": stats['slowest'],
			"cloudpath": cloudpath,
			"chunk_size": chunk_size,
			"dtype": vol.dtype,
			"hostname": socket.gethostname(),
		})

	for chunk_size in CHUNK_SIZES[::-1]:
		cloudpath = 'gs://seunglab-test/test_v0/{}_uncompressed_{}_{}_{}'.format(voltype, *chunk_size)
		vol = CloudVolume(cloudpath, progress=True)
		stats = benchmark(lambda: vol[:], N)

		log({
			"direction": "download",
			"compression": "none",
			"image_type": voltype,
			"N": N,
			"mean": stats['mean'],
			"fastest": stats['fastest'],
			"slowest": stats['slowest'],
			"cloudpath": cloudpath,
			"chunk_size": chunk_size,
			"dtype": vol.dtype,
			"hostname": socket.gethostname(),
		})

benchmark_download('black')
benchmark_download('image')
benchmark_download('segmentation')

benchmark_upload('black')
benchmark_upload('image')
benchmark_upload('segmentation')


logfile.close()



import matplotlib.pyplot as plt
import numpy as np
import csv

import sys
from collections import defaultdict

rows = []

filename = sys.argv[1] 

with open(filename, 'r') as csvfile:
  votereader = csv.reader(csvfile, delimiter='\t', quotechar=None)
  for row in votereader:
        rows.append(row)

headers = rows.pop(0)

stats = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))

for row in rows:
	direction, imgtype, compression, MB, meanMBs = row[1], row[3], row[2], row[6], row[7]
	stats[direction][compression][imgtype][float(MB)] = float(meanMBs)


styles = {
	'black': 'k',
	'image': (.7, .7, .7, 1),
	'segmentation': 'r',
}


index = 0
for direction in sorted(stats.keys()):
	direxpr = stats[direction]

	for compress in sorted(direxpr.keys()):
		imageexper = direxpr[compress]
		index += 1

		plt.subplot(2, 2, index)
		plt.title(filename + ': CloudVolume ' + direction + ' Speed w/ ' + compress + ' Compression')
		plt.xlabel('Chunk Size (MB)')
		plt.ylabel('MB/sec')

		lines = []
		legends = []

		for imgtype in imageexper.keys():
			style = styles[imgtype]
			experiment = imageexper[imgtype]
			xdata = sorted([ float(mb) for mb in experiment.keys() ])
			ydata = [ float(experiment[mb]) for mb in xdata ]
			line = plt.plot(xdata, ydata, color=style, linestyle='-', linewidth=2)
			lines.append(line)
			legends.append(imgtype)

		plt.legend(legends)
plt.show()









from functools import partial

import tempfile
import cloudvolume
import numpy as np
import shutil
import posixpath
import pytest
import os
from scipy import sparse 
import sys

tempdir = tempfile.mkdtemp()
TEST_PATH = "file://{}".format(tempdir)
TEST_DATASET_NAME = "testvol"
PRECOMPUTED_MESH_TEST_DATASET_NAME = "meshvol_precompute"
DRACO_MESH_TEST_DATASET_NAME = "meshvol_draco"
PCG_LOCATION = "https://www.dynamicannotationframework.com/"
TEST_SEG_ID = 144115188084020434

@pytest.fixture()
def cv_graphene_mesh_precomputed(requests_mock):
  test_dir = os.path.dirname(os.path.abspath(__file__))
  graphene_test_cv_dir = os.path.join(test_dir,'test_cv')
  graphene_test_cv_path = "file://{}".format(graphene_test_cv_dir)

  info_d = {
    "data_dir": graphene_test_cv_path,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [
        512,
        512,
        128
      ],
      "n_layers": 9,
      "n_bits_for_layer_id": 8,
      "spatial_bit_masks": {
        x: 10 for x in range(255)
      },
    },
    "chunks_start_at_voxel_offset": False,
    "mesh": "mesh_mip_2_err_40_sv16",
    "num_channels": 1,
    "scales": [{
      "chunk_sizes": [[ 512, 512, 16 ]],
      "compressed_segmentation_block_size": [
        8,
        8,
        8
      ],
      "encoding": "compressed_segmentation",
      "key": "8_8_40",
      "resolution": [
        8,
        8,
        40
      ],
      "size": [
        43520,
        26112,
        2176
      ],
      "voxel_offset": [
        17920,
        14848,
        0
      ]
    }],
    "type": "segmentation"
  }
  requests_mock.get(PCG_LOCATION+PRECOMPUTED_MESH_TEST_DATASET_NAME+"/info", json=info_d)
  
  frag_files = os.listdir(os.path.join(graphene_test_cv_dir, info_d['mesh']))
  # the file are saved as .gz but we want to list the non gz version
  # as cloudvolume will take care of finding the compressed files
  frag_files = [f[:-3] for f in frag_files if f[0]=='9']
  frag_d = {'fragments':frag_files}
  mock_url = PCG_MESH_LOCATION + PRECOMPUTED_MESH_TEST_DATASET_NAME+"/manifest/{}:0?verify=True".format(TEST_SEG_ID)
  requests_mock.get(mock_url, json=frag_d)

  cloudpath = "graphene://{}{}".format(PCG_LOCATION, PRECOMPUTED_MESH_TEST_DATASET_NAME)
  yield cloudvolume.CloudVolume(cloudpath)

@pytest.fixture()
def cv_graphene_mesh_draco(requests_mock):
  test_dir = os.path.dirname(os.path.abspath(__file__))
  graphene_test_cv_dir = os.path.join(test_dir,'test_cv')
  graphene_test_cv_path = "file://{}".format(graphene_test_cv_dir)

  info_d = {
    "data_dir": graphene_test_cv_path,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [
        512,
        512,
        128
      ],
      "n_layers": 9,
      "n_bits_for_layer_id": 8,
      "spatial_bit_masks": {
        x: 10 for x in range(255)
      },
    },
    "chunks_start_at_voxel_offset": False,
    "mesh": "mesh_mip_2_draco_sv16",
    "mesh_metadata": {
      "max_meshed_layer": 6,
      "uniform_draco_grid_size": 21
    },
    "num_channels": 1,
    "scales": [{
      "chunk_sizes": [[ 512, 512, 16 ]],
      "compressed_segmentation_block_size": [
        8,
        8,
        8
      ],
      "encoding": "compressed_segmentation",
      "key": "8_8_40",
      "resolution": [
        8,
        8,
        40
      ],
      "size": [
        43520,
        26112,
        2176
      ],
      "voxel_offset": [
        17920,
        14848,
        0
      ]
    }],
    "type": "segmentation"
  }
  infourl = posixpath.join(PCG_LOCATION, 'meshing/table', DRACO_MESH_TEST_DATASET_NAME, "info")
  requests_mock.get(infourl, json=info_d)
  
  frag_files = os.listdir(os.path.join(graphene_test_cv_dir, info_d['mesh']))
  # we want to filter out the manifest file
  frag_files = [ f for f in frag_files if f[0] == '1' ]
  frag_d = { 'fragments': frag_files }
  mock_url = posixpath.join(
    PCG_LOCATION, 'meshing/api/v1/table', 
    DRACO_MESH_TEST_DATASET_NAME, 
    "manifest/{}:0?verify=True".format(TEST_SEG_ID)
  )
  requests_mock.get(mock_url, json=frag_d)
  
  cloudpath = posixpath.join(PCG_LOCATION, 'meshing/api/v1', DRACO_MESH_TEST_DATASET_NAME)
  yield cloudvolume.CloudVolume('graphene://' + cloudpath)

@pytest.fixture(scope='session')
def cv_supervoxels(N=64, blockN=16):

  block_per_row = int(N / blockN)

  chunk_size = [32, 32, 32]
  info = cloudvolume.CloudVolume.create_new_info(
    num_channels=1,
    layer_type='segmentation',
    data_type='uint64',
    encoding='raw',
    resolution=[4, 4, 40],  # Voxel scaling, units are in nanometers
    voxel_offset=[0, 0, 0],  # x,y,z offset in voxels from the origin
    # Pick a convenient size for your underlying chunk representation
    # Powers of two are recommended, doesn't need to cover image exactly
    chunk_size=chunk_size,  # units are voxels
    volume_size=[N, N, N],
  )

  vol = cloudvolume.CloudVolume(TEST_PATH, info=info)
  vol.commit_info()
  xx, yy, zz = np.meshgrid(*[np.arange(0, N) for cs in chunk_size])
  id_ind = (
    np.uint64(xx / blockN),
    np.uint64(yy / blockN),
    np.uint64(zz / blockN)
  )
  id_shape = (block_per_row, block_per_row, block_per_row)

  seg = np.ravel_multi_index(id_ind, id_shape)
  vol[:] = np.uint64(seg)

  yield TEST_PATH

  shutil.rmtree(tempdir)


@pytest.fixture()
def graphene_vol(cv_supervoxels,  requests_mock, monkeypatch, N=64):

  chunk_size = [32, 32, 32]

  info_d = {
    "data_dir": cv_supervoxels,
    "data_type": "uint64",
    "graph": {
      "chunk_size": [64, 64, 64]
    },
    "mesh": "mesh_mip_2_err_40_sv16",
    "num_channels": 1,
    "scales": [
      {
        "chunk_sizes": [
          [32, 32, 32]
        ],
        "compressed_segmentation_block_size": [8, 8, 8],
        "encoding": "compressed_segmentation",
        "key": "4_4_40",
        "resolution": [4, 4, 40],
        "size": [N, N, N],
        "voxel_offset": [0, 0, 0]
      }
    ],
    "type": "segmentation"
  }

  infourl = posixpath.join(PCG_LOCATION, 'segmentation/table', TEST_DATASET_NAME, "info")
  requests_mock.get(infourl, json=info_d)
  
  def mock_get_leaves(self, root_id, bbox, mip):
    return np.array([0,1,2,3], dtype=np.uint64)

  cloudpath = "graphene://" + posixpath.join(PCG_LOCATION, 'segmentation', 'api/v1/', TEST_DATASET_NAME)

  gcv = cloudvolume.CloudVolume(cloudpath)
  gcv.get_leaves = partial(mock_get_leaves, gcv)
  yield gcv

def test_gcv(graphene_vol):
  cutout = graphene_vol.download(np.s_[0:5,0:5,0:5], segids=[999])
  assert (np.all(cutout==999))
  cutout_sv = graphene_vol[0:5,0:5,0:5]
  assert cutout_sv.shape == (5,5,5,1)
  assert graphene_vol[0,0,0].shape == (1,1,1,1)


def faces_to_edges(faces, return_index=False):
  """
  Given a list of faces (n,3), return a list of edges (n*3,2)
  Parameters
  -----------
  faces : (n, 3) int
    Vertex indices representing faces
  Returns
  -----------
  edges : (n*3, 2) int
    Vertex indices representing edges
  """
  faces = np.asanyarray(faces)

  # each face has three edges
  edges = faces[:, [0, 1, 1, 2, 2, 0]].reshape((-1, 2))

  if return_index:
    # edges are in order of faces due to reshape
    face_index = np.tile(np.arange(len(faces)),
               (3, 1)).T.reshape(-1)
    return edges, face_index
  return edges

  
def create_csgraph(vertices, edges, euclidean_weight=True, directed=False):
  '''
  Builds a csr graph from vertices and edges, with optional control
  over weights as boolean or based on Euclidean distance.
  '''
  if euclidean_weight:
    xs = vertices[edges[:,0]]
    ys = vertices[edges[:,1]]
    weights = np.linalg.norm(xs-ys, axis=1)
    use_dtype = np.float32
  else:   
    weights = np.ones((len(edges),)).astype(np.int8)
    use_dtype = np.int8 

  if directed:
    edges = edges.T
  else:
    edges = np.concatenate([edges.T, edges.T[[1, 0]]], axis=1)
    weights = np.concatenate([weights, weights]).astype(dtype=use_dtype)

  csgraph = sparse.csr_matrix((weights, edges),
                shape=[len(vertices), ] * 2,
                dtype=use_dtype)

  return csgraph

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_decode_segid(cv_graphene_mesh_draco):
  decoded = cv_graphene_mesh_draco.meta.decode_label(648518346349515986)
  assert decoded.level == 9
  assert decoded.x == 0
  assert decoded.y == 0
  assert decoded.z == 0
  assert decoded.segid == 8164562

  level = '00000101' # 5
  x = '0000000001' # 1 
  y = '0000000010' # 2 
  z = '0000000011' # 3
  segid = '00000000000000000000001010' # 10

  label = int(level + x + y + z + segid, 2)
  decoded = cv_graphene_mesh_draco.meta.decode_label(label)
  
  assert decoded.level == 5
  assert decoded.x == 1
  assert decoded.y == 2
  assert decoded.z == 3
  assert decoded.segid == 10

  encoded = cv_graphene_mesh_draco.meta.encode_label(*decoded)
  assert decoded == cv_graphene_mesh_draco.meta.decode_label(encoded)


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_graphene_mesh_get(cv_graphene_mesh_precomputed):

  mesh = cv_graphene_mesh_precomputed.mesh.get(TEST_SEG_ID)
  edges = faces_to_edges(mesh[TEST_SEG_ID].faces)
  graph = create_csgraph(mesh[TEST_SEG_ID].vertices,
               edges,
               directed=False)
  ccs, labels =  sparse.csgraph.connected_components(graph,
                             directed=False)
  assert(ccs==3)

@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
def test_graphene_mesh_get(cv_graphene_mesh_draco):

  mesh = cv_graphene_mesh_draco.mesh.get(TEST_SEG_ID)
  edges = faces_to_edges(mesh[TEST_SEG_ID].faces)
  graph = create_csgraph(mesh[TEST_SEG_ID].vertices,
               edges,
               directed=False)
  ccs, labels =  sparse.csgraph.connected_components(graph,
                             directed=False)
  assert(ccs==3)




from __future__ import print_function

from collections import namedtuple, defaultdict
import copy
import json
from os.path import basename
import struct

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from . import mmh3
from ... import compression
from ...lib import jsonify, toiter
from ...lru import LRU
from ...exceptions import SpecViolation, EmptyFileException

ShardLocation = namedtuple('ShardLocation', 
  ('shard_number', 'minishard_number', 'remainder')
)

uint64 = np.uint64

class ShardingSpecification(object):
  def __init__(
    self, type, preshift_bits, 
    hash, minishard_bits, 
    shard_bits, 
    minishard_index_encoding='raw', 
    data_encoding='raw'
  ):

    self.type = type 
    self.preshift_bits = uint64(preshift_bits)
    self.hash = hash 
    self.minishard_bits = uint64(minishard_bits)
    self.shard_bits = uint64(shard_bits)
    self.minishard_index_encoding = minishard_index_encoding
    self.data_encoding = data_encoding

    self.minishard_mask = self.compute_minishard_mask(self.minishard_bits)
    self.shard_mask = self.compute_shard_mask(self.shard_bits, self.minishard_bits)              

    self.validate()

  def clone(self):
    return ShardingSpecification.from_dict(self.to_dict())

  def index_length(self):
    return int((2 ** self.minishard_bits) * 16)

  @property
  def hash(self):
    return self._hash

  @hash.setter
  def hash(self, val):
    if val == 'identity':
      self.hashfn = lambda x: uint64(x)
    elif val == 'murmurhash3_x86_128':
      self.hashfn = lambda x: uint64(mmh3.hash64(uint64(x).tobytes(), x64arch=False)[0]) 
    else:
      raise SpecViolation("hash {} must be either 'identity' or 'murmurhash3_x86_128'".format(val))

    self._hash = val

  @property
  def preshift_bits(self):
    return self._preshift_bits
  
  @preshift_bits.setter
  def preshift_bits(self, val):
    self._preshift_bits = uint64(val) 

  @property
  def shard_bits(self):
    return self._shard_bits
  
  @shard_bits.setter
  def shard_bits(self, val):
    self._shard_bits = uint64(val) 

  @property
  def minishard_bits(self):
    return self._minishard_bits
  
  @minishard_bits.setter
  def minishard_bits(self, val):
    val = uint64(val)
    self.minishard_mask = self.compute_minishard_mask(val)
    self._minishard_bits = uint64(val)

  def compute_minishard_mask(self, val):
    if val < 0:
      raise ValueError(str(val) + " must be greater or equal to than zero.")
    elif val == 0:
      return uint64(0)

    minishard_mask = uint64(1)
    for i in range(val - uint64(1)):
      minishard_mask <<= uint64(1)
      minishard_mask |= uint64(1)
    return uint64(minishard_mask)

  def compute_shard_mask(self, shard_bits, minishard_bits):
    ones64 = uint64(0xffffffffffffffff)
    movement = uint64(minishard_bits + shard_bits)
    shard_mask = ~((ones64 >> movement) << movement)
    minishard_mask = self.compute_minishard_mask(minishard_bits)
    return shard_mask & (~minishard_mask)

  @classmethod
  def from_json(cls, vals):
    dct = json.loads(vals.decode('utf8'))
    return cls.from_dict(dct)

  def to_json(self):
    return jsonify(self.to_dict())

  @classmethod
  def from_dict(cls, vals):
    vals = copy.deepcopy(vals)
    vals['type'] = vals['@type']
    del vals['@type']
    return cls(**vals)

  def to_dict(self):
    return {
      '@type': self.type,
      'preshift_bits': self.preshift_bits,
      'hash': self.hash,
      'minishard_bits': self.minishard_bits,
      'shard_bits': self.shard_bits,
      'minishard_index_encoding': self.minishard_index_encoding,
      'data_encoding': self.data_encoding,
    }

  def compute_shard_location(self, key):
    chunkid = uint64(key) >> uint64(self.preshift_bits)
    chunkid = self.hashfn(chunkid)
    minishard_number = uint64(chunkid & self.minishard_mask)
    shard_number = uint64((chunkid & self.shard_mask) >> uint64(self.minishard_bits))
    shard_number = format(shard_number, 'x').zfill(int(np.ceil(self.shard_bits / 4.0)))
    remainder = chunkid >> uint64(self.minishard_bits + self.shard_bits)

    return ShardLocation(shard_number, minishard_number, remainder)

  def synthesize_shards(self, data, progress=False):
    """
    Given this specification and a comprehensive listing of
    all the items that could be combined into a given shard,
    synthesize the shard files for this set of labels.

    data: { label: binary, ... }

    e.g. { 5: b'...', 7: b'...' }

    Returns: {
      $filename: binary data,
    }
    """
    return synthesize_shard_files(self, data, progress)

  def synthesize_shard(self, labels, progress=False, presorted=False):
    """
    Assemble a shard file from a group of labels that all belong in the same shard.

    Assembles the .shard file like:
    [ shard index; minishards; all minishard indices ]

    label_group: 
      If presorted is True:
        { minishardno: { label: binary, ... }, ... }
      If presorted is False:
        { label: binary }
    progress: show progress bars

    Returns: binary representing a shard file 
    """
    return synthesize_shard_file(self, labels, progress, presorted)

  def validate(self):
    if self.type not in ('neuroglancer_uint64_sharded_v1',):
      raise SpecViolation(
        "@type ({}) must be 'neuroglancer_uint64_sharded_v1'." \
        .format(self.type)
      )

    if not (64 > self.preshift_bits >= 0):
      raise SpecViolation("preshift_bits must be a whole number less than 64: {}".format(self.preshift_bits))

    if not (64 >= self.minishard_bits >= 0):
      raise SpecViolation("minishard_bits must be between 0 and 64 inclusive: {}".format(self.minishard_bits))

    if not (64 >= self.shard_bits >= 0):
      raise SpecViolation("shard_bits must be between 0 and 64 inclusive: {}".format(self.shard_bits))

    if self.minishard_bits + self.shard_bits > 64:
      raise SpecViolation(
        "minishard_bits and shard_bits must sum to less than or equal to 64: minishard_bits<{}> + shard_bits<{}> = {}".format(
        self.minishard_bits, self.shard_bits, self.minishard_bits + self.shard_bits
      ))

    if self.hash not in ('identity', 'murmurhash3_x86_128'):
      raise SpecViolation("hash {} must be either 'identity' or 'murmurhash3_x86_128'".format(self.hash))

    if self.minishard_index_encoding not in ('raw', 'gzip'):
      raise SpecViolation("minishard_index_encoding only supports values 'raw' or 'gzip'.")

    if self.data_encoding not in ('raw', 'gzip'):
      raise SpecViolation("data_encoding only supports values 'raw' or 'gzip'.")
    
  def __str__(self):
    return "ShardingSpecification::" + str(self.to_dict())

class ShardReader(object):
  def __init__(
    self, meta, cache, spec,
    shard_index_cache_size=512,
    minishard_index_cache_size=128,
    green=False
  ):
    """
    Reads standard Precomputed shard files. 

    meta: a PrecomputedMetadata class
    cache: a CacheService instance
    spec: a ShardingSpecification instance

    shard_index_cache_size: size of LRU cache for fixed indices 
    minishard_index_cache_size: size of LRU cache for minishard indices
    """
    self.meta = meta
    self.cache = cache
    self.spec = spec
    self.green = green

    self.shard_index_cache = LRU(shard_index_cache_size)
    self.minishard_index_cache = LRU(minishard_index_cache_size)

  def get_filename(self, label):
    return self.compute_shard_location(label)[0]

  def compute_shard_location(self, label):
    """
    Returns (filename, shard_number) for meshes and skeletons. 
    Images require a different scheme.
    """
    shard_loc = self.spec.compute_shard_location(label)
    filename = str(shard_loc.shard_number) + '.shard'
    return (filename, shard_loc.minishard_number)

  def get_index(self, filename, path=""):
    """
    Retrieves the shard index which is used for 
    locating the appropriate minishard index.

    Returns: 2^minishard_bits entries of a uint64 
      array of [[ byte start, byte end ], ... ] 
    """
    indices = self.get_indices([ filename ], path, progress=False)
    index = list(indices.values())[0]
    if index is None:
      raise EmptyFileException(filename + " was zero bytes.")
    return index

  def get_indices(self, filenames, path="", progress=None):
    """
    For all given files, retrieves the shard index which 
    is used for locating the appropriate minishard indices.

    Returns: { 
      path_to_/filename.shard: 2^minishard_bits entries of a uint64 
            array of [[ byte start, byte end ], ... ],
      ...
    } 
    """
    filenames = toiter(filenames)
    filenames = [ self.meta.join(path, fname) for fname in filenames ]
    fufilled = { 
      fname: self.shard_index_cache[fname] \
      for fname in filenames \
      if fname in self.shard_index_cache  
    }

    requests = []
    for fname in filenames:
      if fname in fufilled:
        continue
      requests.append({
        'path': fname, 
        'local_alias': fname + '.index', 
        'start': 0, 
        'end': self.spec.index_length(),
      })

    progress = 'Shard Indices' if progress else False
    binaries = self.cache.download_as(requests, progress=progress)
    for (fname, start, end), content in binaries.items():
      try:
        index = self.decode_index(content, fname)
        self.shard_index_cache[fname] = index
        fufilled[fname] = index
      except EmptyFileException:
        self.shard_index_cache[fname] = None
        fufilled[fname] = None

    return fufilled

  def decode_index(self, binary, filename='Shard'):
    if binary is None or len(binary) == 0:
      raise EmptyFileException(filename + " was zero bytes.")
    elif len(binary) != self.spec.index_length():
      raise SpecViolation(
        filename + ": shard index was an incorrect length ({}) for this specification ({}).".format(
          len(binary), self.spec.index_length()
        ))
    
    index = np.frombuffer(binary, dtype=np.uint64)
    index = index.reshape( (index.size // 2, 2), order='C' )
    return index + self.spec.index_length()

  def decode_minishard_index(self, minishard_index, filename=''):
    """Returns [[label, offset, size], ... ] where offset and size are in bytes."""

    if self.spec.minishard_index_encoding != 'raw':
      minishard_index = compression.decompress(
        minishard_index, encoding=self.spec.minishard_index_encoding, filename=filename
      )

    minishard_index = np.copy(np.frombuffer(minishard_index, dtype=np.uint64))
    minishard_index = minishard_index.reshape( (3, len(minishard_index) // 3), order='C' ).T

    for i in range(1, minishard_index.shape[0]):
      minishard_index[i, 0] += minishard_index[i-1, 0]
      minishard_index[i, 1] += minishard_index[i-1, 1] + minishard_index[i-1, 2]

    minishard_index[:,1] += self.spec.index_length()

    return minishard_index 

  def get_minishard_index(self, filename, index, minishard_no, path=""):
    """
    Retrieves the minishard index for a given minishard number.

    Returns: uint64 Nx3 array with multiple rows of [segid, byte start, byte end]
    """
    res = self.get_minishard_indices(filename, index, minishard_no, path)
    return res[minishard_no]

  def get_minishard_indices(self, filename, index, minishard_nos, path=""):
    """
    Retrieves the minishard indices for a set of minishard numbers.

    Returns: { minishard_no: uint64 Nx3 array of [segid, byte start, byte end], ... }
    """
    res = self.get_minishard_indices_for_files(( (filename, index, minishard_nos), ), path)
    return res[filename]

  def get_minishard_indices_for_files(self, requests, path="", progress=None):
    """
    Fetches the specified minishard indices for all the specified files
    at once. This is required to get high performance as opposed to fetching
    the all minishard indices for a single file.

    requests: iterable of tuples
      [  (filename, index, minishard_numbers), ... ]

    Returns: map of filename -> minishard numbers -> minishard indices

    e.g. 
    {
      filename_1: {
          0: uint64 Nx3 array of [segid, byte start, byte end],
          1: ...,
      }
      filename_2: ...
    }
    """
    fufilled_by_filename = defaultdict(dict)
    msn_map = {}

    download_requests = []
    for filename, index, minishard_nos in requests:
      fufilled_requests, pending_requests = self.compute_minishard_index_requests(
        filename, index, minishard_nos, path
      ) 
      fufilled_by_filename[filename] = fufilled_requests
      for msn, start, end in pending_requests:
        msn_map[(filename, start, end)] = msn

        filepath = self.meta.join(path, filename)

        download_requests.append({
          'path': filepath,
          'local_alias': '{}-{}.msi'.format(filepath, msn),
          'start': start,
          'end': end,
        })

    progress = 'Minishard Indices' if progress else False
    results = self.cache.download_as(download_requests, progress=progress)
  
    for (filename, start, end), content in results.items():
      filename = basename(filename)
      cache_key = (filename, start, end)
      msn = msn_map[cache_key]
      minishard_index = self.decode_minishard_index(content, filename)
      self.minishard_index_cache[cache_key] = minishard_index
      fufilled_by_filename[filename][msn] = minishard_index

    return fufilled_by_filename

  def compute_minishard_index_requests(self, filename, index, minishard_nos, path=""):
    """
    Helper method for get_minishard_indices_for_files. 
    Computes which requests must be made over the network vs can be fufilled from LRU cache.
    """
    minishard_nos = toiter(minishard_nos)

    if index is None:
      return ({ msn: None for msn in minishard_nos }, [])

    fufilled_requests = {}

    byte_ranges = {}
    for msn in minishard_nos:
      bytes_start, bytes_end = index[msn]

      # most typically: [0,0] for an incomplete shard
      if bytes_start == bytes_end:
        fufilled_requests[msn] = None
        continue

      bytes_start, bytes_end = int(bytes_start), int(bytes_end)
      byte_ranges[msn] = (bytes_start, bytes_end)

    full_path = self.meta.join(self.meta.cloudpath, path)

    pending_requests = []
    for msn, (bytes_start, bytes_end) in byte_ranges.items():
      cache_key = (filename, bytes_start, bytes_end)
      if cache_key in self.minishard_index_cache:
        fufilled_requests[msn] = self.minishard_index_cache[cache_key]
      else:
        pending_requests.append((msn, bytes_start, bytes_end))

    return (fufilled_requests, pending_requests)

  def exists(self, labels, path="", return_byte_range=False, progress=None):
    """
    Checks a shard's minishard index for whether a file exists.

    If return_byte_range = False:
      OUTPUT = SHARD_FILEPATH or None if not exists
    Else:
      OUTPUT = [ SHARD_FILEPATH or None, byte_start, num_bytes ]

    Returns:
      If labels is not an iterable:
        return OUTPUT
      Else:
        return { label_1: OUTPUT, label_2: OUTPUT, ... }
    """
    return_one = False

    try:
      iter(labels)
    except TypeError:
      return_one = True

    to_labels = defaultdict(list)
    to_all_labels = defaultdict(list)
    filename_to_minishard_num = defaultdict(list)

    for label in set(toiter(labels)):
      filename, minishard_number = self.compute_shard_location(label)
      to_labels[(filename, minishard_number)].append(label)
      to_all_labels[filename].append(label)
      filename_to_minishard_num[filename].append(minishard_number)

    indices = self.get_indices(to_all_labels.keys(), path, progress=progress)

    all_minishards = self.get_minishard_indices_for_files([ 
      (basename(filepath), index, filename_to_minishard_num[basename(filepath)]) \
      for filepath, index in indices.items()
    ], path, progress=progress)

    results = {}
    for filename, file_minishards in all_minishards.items():
      filepath = self.meta.join(path, filename)
      for mini_no, msi in file_minishards.items():
        labels = to_labels[(filename, mini_no)]

        for label in labels:
          if msi is None:
            results[label] = None
            continue

          idx = np.where(msi[:,0] == label)[0]
          if len(idx) == 0:
            results[label] = None
          else:
            if return_byte_range:
              _, offset, size = msi[idx,:][0]
              results[label] = [ filepath, int(offset), int(size) ]
            else:
              results[label] = filepath

    if return_one:
      return(list(results.values())[0])
    return results

  def disassemble_shard(self, shard):
    """
    Given an entire shard as a bytestring, convert 
    it into a dict of { label: byte content }.
    """
    index = self.decode_index(shard[:self.spec.index_length()])
    shattered = {}
    for start, end in index:
      start, end = int(start), int(end)
      if start == end:
        continue

      msi = self.decode_minishard_index(shard[start:end])
      for label, offset, size in msi:
        offset, size = int(offset), int(size)
        binary = shard[offset:offset+size]
        
        if self.spec.data_encoding != 'raw':
          binary = compression.decompress(binary, encoding=self.spec.data_encoding)
        
        shattered[label] = binary

    return shattered

  def get_data(self, label, path="", progress=None):
    """Fetches data from shards.

    label: one or more segment ids
    path: subdirectory path
    progress: display progress bars

    Return: 
      if label is a scalar:
        a byte string
      else: (label is an iterable)
        {
          label_1: byte string,
          ....
        }
    """
    label, return_multiple = toiter(label, is_iter=True)
    label = set(( int(l) for l in label))

    results = {}
    if self.cache.enabled:
      results = self.cache.get([ 
        self.meta.join(path, str(lbl)) for lbl in label
      ], progress=progress)

    for cloudpath, content in results.items():
      if content is None:
        label.remove(int(basename(cloudpath)))


    # { label: [ filename, byte start, num_bytes ] }
    exists = self.exists(label, path, return_byte_range=True, progress=progress)
    for k in list(exists.keys()):
      if exists[k] is None:
        results[k] = None
        del exists[k]

    key_label = { (basename(v[0]), v[1], v[2]): k for k,v in exists.items() }

    files = ( 
      { 'path': basename(ext[0]), 'start': int(ext[1]), 'end': int(ext[1]) + int(ext[2]) }
      for ext in exists.values()
    )

    full_path = self.meta.join(self.meta.cloudpath, path)
    remote_files = CloudFiles(full_path, progress=progress, green=self.green).get(files)

    binaries = {}
    for res in remote_files:
      if res['error']:
        raise res['error']
      start, end = res['byte_range']
      key = (res['path'], start, end - start)
      lbl = key_label[key]
      binaries[lbl] = res['content']
    del remote_files

    if self.spec.data_encoding != 'raw':
      for filepath, binary in tqdm(binaries.items(), desc="Decompressing", disable=(not progress)):
        if binary is None:
          continue
        binaries[filepath] = compression.decompress(
          binary, encoding=self.spec.data_encoding, filename=filepath
        )
      
    if self.cache.enabled:
      self.cache.put([ 
        (filepath, binary) for filepath, binary in binaries.items()
      ], progress=progress)

    results.update(binaries)

    if return_multiple:
      return results
    return next(iter(results.values()))

  def list_labels(self, filename, path="", size=False):
    """
    List all the labels in the index of a given shard file.

    size: (bool) if True, list the size in bytes of each label

    Returns: 
      if not size:
        np.uint64 array of labels 
      else:
        [ (label, size in bytes), ... ] in descending order of size
    """
    index = self.get_index(filename, path)
    all_minishard_nos = list(range(len(index)))
    minishard_indicies = self.get_minishard_indices(filename, index, all_minishard_nos, path)
    minishard_indicies = [  
      msi for msi in minishard_indicies.values() if msi is not None
    ]
    if not size:
      labels = np.concatenate([  
        msi[:,0] for msi in minishard_indicies
      ])
      return np.sort(labels)
    else:
      labels = np.concatenate([  
        msi[:,:]
        for msi in minishard_indicies
      ])
      labels = [ (row[0], row[2]) for row in labels[:] ]
      return sorted(labels, key=lambda x: x[1], reverse=True)

def synthesize_shard_files(spec, data, progress=False):
  """
  From a set of data guaranteed to constitute one or more
  complete and comprehensive shards (no partial shards) 
  return a set of files ready for upload.

  WARNING: This function is only appropriate for Precomputed
  meshes and skeletons. Use the synthesize_shard_file (singular)
  function to create arbitrarily named and assigned shard files.

  spec: a ShardingSpecification
  data: { label: binary, ... }

  Returns: { filename: binary, ... }
  """
  shard_groupings = defaultdict(lambda: defaultdict(dict))
  pbar = tqdm(
    data.items(), 
    desc='Creating Shard Groupings', 
    disable=(not progress)
  )

  for label, binary in pbar:
    loc = spec.compute_shard_location(label)
    shard_groupings[loc.shard_number][loc.minishard_number][label] = binary

  shard_files = {}

  pbar = tqdm(
    shard_groupings.items(), 
    desc="Synthesizing Shard Files", 
    disable=(not progress)
  )

  for shardno, shardgrp in pbar:
    filename = str(shardno) + '.shard'
    shard_files[filename] = synthesize_shard_file(spec, shardgrp, progress=(progress > 1), presorted=True)

  return shard_files

# NB: This is going to be memory hungry and can be optimized
def synthesize_shard_file(spec, label_group, progress=False, presorted=False):
  """
  Assemble a shard file from a group of labels that all belong in the same shard.

  Assembles the .shard file like:
  [ shard index; minishards; all minishard indices ]

  spec: ShardingSpecification
  label_group: 
    If presorted is True:
      { minishardno: { label: binary, ... }, ... }
    If presorted is False:
      { label: binary }
  progress: show progress bars

  Returns: binary representing a shard file
  """
  minishardnos = []
  minishard_indicies = []
  minishards = []

  if presorted:
    minishard_mapping = label_group
  else:
    minishard_mapping = defaultdict(dict)
    pbar = tqdm(label_group.items(), disable=(not progress), desc="Assigning Minishards")
    for label, binary in pbar:
      loc = spec.compute_shard_location(label)
      minishard_mapping[loc.minishard_number][label] = binary

  del label_group

  for minishardno, minishardgrp in tqdm(minishard_mapping.items(), desc="Minishard Indices", disable=(not progress)):
    labels = sorted([ int(label) for label in minishardgrp.keys() ])
    if len(labels) == 0:
      continue

    minishard_index = np.zeros( (3, len(labels)), dtype=np.uint64, order='C')
    minishard = b''
    
    # label and offset are delta encoded
    last_label = 0
    for i, label in enumerate(labels):
      binary = minishardgrp[label]
      if spec.data_encoding != 'raw':
        binary = compression.compress(binary, method=spec.data_encoding)

      minishard_index[0, i] = label - last_label
      minishard_index[1, i] = 0 # minishard_index[2, i - 1]
      minishard_index[2, i] = len(binary)
      minishard += binary
      last_label = label
      del minishardgrp[label]
    
    minishardnos.append(minishardno)
    minishard_indicies.append(minishard_index) 
    minishards.append(minishard)

  del minishard_mapping

  cum_minishard_size = 0
  for idx, minishard in zip(minishard_indicies, minishards):
    idx[1, 0] = cum_minishard_size
    cum_minishard_size += len(minishard)

  if progress:
    print("Partial assembly of minishard indicies and data... ", end="", flush=True)

  variable_index_part = [ idx.tobytes('C') for idx in minishard_indicies ]
  if spec.minishard_index_encoding != 'raw':
    variable_index_part = [ 
      compression.compress(idx, method=spec.minishard_index_encoding) \
      for idx in variable_index_part 
    ]

  data_part = b''.join(minishards)
  del minishards

  if progress:
    print("Assembled.")

  fixed_index = np.zeros( 
    (int(2 ** spec.minishard_bits), 2), 
    dtype=np.uint64, order='C'
  )

  start = len(data_part)
  end = len(data_part)
  for i, idx in zip(minishardnos, variable_index_part):
    start = end
    end += len(idx)
    fixed_index[i, 0] = start
    fixed_index[i, 1] = end

  if progress:
    print("Final assembly... ", end="", flush=True)

  # The order here is important. The fixed index must go first because the locations
  # of the other parts are calculated with it implicitly in front. The variable
  # index must go last because otherwise compressing it will affect offset of the
  # data it is attempting to index.

  result = fixed_index.tobytes('C') + data_part + b''.join(variable_index_part) 

  if progress:
    print("Done.")

  return result



import pytest

import shutil
import numpy as np

import os

from cloudvolume import Storage, CloudVolume
from cloudvolume.lib import Bbox, Vec, min2, max2, find_closest_divisor

TEST_NUMBER = np.random.randint(0, 999999)
CLOUD_BUCKET = 'seunglab-test'

layer_path = '/tmp/removeme/'

def create_storage(layer_name='layer'):
    stor_path = os.path.join(layer_path, layer_name)
    return Storage('file://' + stor_path, n_threads=0)

def create_image(size, layer_type="image", dtype=None):
    default = lambda dt: dtype or dt

    if layer_type == "image":
        return np.random.randint(255, size=size, dtype=default(np.uint8))
    elif layer_type == 'affinities':
        return np.random.uniform(low=0, high=1, size=size).astype(default(np.float32))
    elif layer_type == "segmentation":
        return np.random.randint(0xFFFFFF, size=size, dtype=np.uint32)
    else:
        high = np.array([0], dtype=default(np.uint32)) - 1
        return np.random.randint(high[0], size=size, dtype=default(np.uint32))

def create_layer(size, offset, layer_type="image", layer_name='layer', dtype=None):
    random_data = create_image(size, layer_type, dtype)
    vol = upload_image(random_data, offset, layer_type, layer_name)
    return vol, random_data

def upload_image(image, offset, layer_type, layer_name):
    lpath = 'file://{}'.format(os.path.join(layer_path, layer_name))
    
    neuroglancer_chunk_size = find_closest_divisor(image.shape[:3], closest_to=[64,64,64])

    # Jpeg encoding is lossy so it won't work
    vol = CloudVolume.from_numpy(
      image, 
      vol_path=lpath,
      resolution=(1,1,1), 
      voxel_offset=offset, 
      chunk_size=neuroglancer_chunk_size, 
      layer_type=layer_type, 
      encoding='raw', 
    )
    
    return vol

def delete_layer(path=layer_path):
    if os.path.exists(path):
        shutil.rmtree(path)  



import sys
import time

import multiprocessing as mp
import numpy as np

from .exceptions import UnsupportedFormatError
from .lib import generate_random_string
from .paths import strict_extract, to_https_protocol

# NOTE: Plugins are registered in __init__.py

# Set the interpreter bool
try:
  INTERACTIVE = bool(sys.ps1)
except AttributeError:
  INTERACTIVE = bool(sys.flags.interactive)

REGISTERED_PLUGINS = {}
def register_plugin(key, creation_function):
  REGISTERED_PLUGINS[key.lower()] = creation_function

class SharedConfiguration(object):
  """
  Hack around python's inability to
  pass primatives by reference.
  We would like updates to e.g. mip or parallel
  to be passively absorbed by all listening
  data sources rather than work hard to actively
  synchronize them.
  """
  def __init__(
    self, cdn_cache, compress, compress_level, green,
    mip, parallel, progress,
    *args, **kwargs
  ):
    if type(parallel) == bool:
      parallel = mp.cpu_count() if parallel == True else 1
    elif parallel <= 0:
      raise ValueError('Number of processes must be >= 1. Got: ' + str(parallel))
    else:
      parallel = int(parallel)

    self.cdn_cache = cdn_cache
    self.compress = compress
    self.compress_level = compress_level
    self.green = bool(green)
    self.mip = mip
    self.parallel = parallel 
    self.progress = bool(progress)
    self.args = args
    self.kwargs = kwargs

class CloudVolume(object):
  def __new__(cls,
    cloudpath, mip=0, bounded=True, autocrop=False,
    fill_missing=False, cache=False, compress_cache=None,
    cdn_cache=True, progress=INTERACTIVE, info=None, provenance=None,
    compress=None, compress_level=None, non_aligned_writes=False, parallel=1,
    delete_black_uploads=False, background_color=0,
    green_threads=False, use_https=False,
    max_redirects=10, mesh_dir=None, skel_dir=None
  ):
    """
    A "serverless" Python client for reading and writing arbitrarily large 
    Neuroglancer Precomputed volumes both locally and on cloud services using.  
    A CloudVolume instance represents a dataset interface at a given mip level 
    (it doesn't load the entire dataset into memory).  

    Neuroglancer datasets specify metadata in an `info` file located at the root 
    of a data layer. It contains, among other things, the bounds of the 
    volume described as a 3D "voxel_offset" and 3D "size" in voxels, and the
    resolution of the dataset.

    Example:

      from cloudvolume import CloudVolume

      vol = CloudVolume('gs://mylab/mouse/image', progress=True)
      image = vol[:,:,:] # Download an image stack as a numpy array
      vol[:,:,:] = image # Upload an image stack from a numpy array
      
      label = 1
      mesh = vol.mesh.get(label) 
      skel = vol.skeletons.get(label)

    Required:
      cloudpath: Path to the dataset layer. This should match storage's supported
        providers.

        e.g. Google: gs://$BUCKET/$DATASET/$LAYER/
             S3    : s3://$BUCKET/$DATASET/$LAYER/
             Lcl FS: file:///tmp/$DATASET/$LAYER/
             Boss  : boss://$COLLECTION/$EXPERIMENT/$CHANNEL
             HTTP/S: http(s)://.../$CHANNEL
             matrix: matrix://$BUCKET/$DATASET/$LAYER/
    Optional:
      autocrop: (bool) If the specified retrieval bounding box exceeds the
          volume bounds, process only the area contained inside the volume. 
          This can be useful way to ensure that you are staying inside the 
          bounds when `bounded=False`.
      background_color: (number) Specifies what the "background value" of the
        volume is (traditionally 0). This is mainly for changing the behavior
        of delete_black_uploads.
      bounded: (bool) If a region outside of volume bounds is accessed:
          True: Throw an error
          False: Allow accessing the region. If no files are present, an error 
              will still be thrown. Consider combining this option with 
              `fill_missing=True`. However, this can be dangrous as it allows
              missing files and potentially network errors to be intepreted as 
              zeros.
      cache: (bool or str) Store downs and uploads in a cache on disk
            and preferentially read from it before redownloading.
          - falsey value: no caching will occur.
          - True: cache will be located in a standard location.
          - non-empty string: cache is located at this file path

          After initialization, you can adjust this setting via:
          `cv.cache.enabled = ...` which accepts the same values.

      cdn_cache: (int, bool, or str) Sets Cache-Control HTTP header on uploaded 
        image files. Most cloud providers perform some kind of caching. As of 
        this writing, Google defaults to 3600 seconds. Most of the time you'll 
        want to go with the default. 
        - int: number of seconds for cache to be considered fresh (max-age)
        - bool: True: max-age=3600, False: no-cache
        - str: set the header manually
      compress: (bool, str, None) pick which compression method to use.
          None: (default) gzip for raw arrays and no additional compression
            for compressed_segmentation and fpzip.
          bool: 
            True=gzip, 
            False=no compression, Overrides defaults
          str: 
            'gzip': Extension so that we can add additional methods in the future 
                    like lz4 or zstd. 
            'br': Brotli compression, better compression rate than gzip
            '': no compression (same as False).
      compress_level: (int, None) level for compression. Higher number results
          in better compression but takes longer.
        Defaults to 9 for gzip (ranges from 0 to 9).
        Defaults to 5 for brotli (ranges from 0 to 11).
      compress_cache: (None or bool) If not None, override default compression 
          behavior for the cache.
      delete_black_uploads: (bool) If True, on uploading an entirely black chunk,
          issue a DELETE request instead of a PUT. This can be useful for avoiding storing
          tiny files in the region around an ROI. Some storage systems using erasure coding 
          don't do well with tiny file sizes.
      fill_missing: (bool) If a chunk file is unable to be fetched:
          True: Use a block of zeros
          False: Throw an error
      green_threads: (bool) Use green threads instead of preemptive threads. This
        can result in higher download performance for some compression types. Preemptive
        threads seem to reduce performance on multi-core machines that aren't densely
        loaded as the CPython threads are assigned to multiple cores and the thrashing
        + GIL reduces performance. You'll need to add the following code to the top
        of your program to use green threads:

            import gevent.monkey
            gevent.monkey.patch_all(threads=False)

      info: (dict) In lieu of fetching a neuroglancer info file, use this one.
          This is useful when creating new datasets and for repeatedly initializing
          a new cloudvolume instance.
      max_redirects: (int) if > 0, allow up to this many redirects via info file 'redirect'
          data fields. If <= 0, allow no redirections and access the current info file directly
          without raising an error.
      mesh_dir: (str) if not None, override the info['mesh'] key before pulling the
        mesh info file.
      mip: (int or iterable) Which level of downsampling to read and write from.
          0 is the highest resolution. You can also specify the voxel resolution
          like mip=[6,6,30] which will search for the appropriate mip level.
      non_aligned_writes: (bool) Enable non-aligned writes. Not multiprocessing 
          safe without careful design. When not enabled, a 
          cloudvolume.exceptions.AlignmentError is thrown for non-aligned writes. 
          
          https://github.com/seung-lab/cloud-volume/wiki/Advanced-Topic:-Non-Aligned-Writes

      parallel (int: 1, bool): Number of extra processes to launch, 1 means only 
          use the main process. If parallel is True use the number of CPUs 
          returned by multiprocessing.cpu_count(). When parallel > 1, shared
          memory (Linux) or emulated shared memory via files (other platforms) 
          is used by the underlying download.
      progress: (bool) Show progress bars. 
          Defaults to True in interactive python, False in script execution mode.
      provenance: (string, dict) In lieu of fetching a provenance 
          file, use this one. 
      skel_dir: (str) if not None, override the info['skeletons'] key before 
        pulling the skeleton info file.
      use_https: (bool) maps gs:// and s3:// to their respective https paths. The 
        https paths hit a cached, read-only version of the data and may be faster.
    """
    if use_https:
      cloudpath = to_https_protocol(cloudpath)

    kwargs = dict(locals())
    del kwargs['cls']

    path = strict_extract(cloudpath)
    if path.format in REGISTERED_PLUGINS:
      return REGISTERED_PLUGINS[path.format](**kwargs)
    else:
      raise UnsupportedFormatError(
        "Unknown format {}".format(path.format)
      )

  @classmethod
  def create_new_info(cls, *args, **kwargs):
    from .frontends import CloudVolumePrecomputed
    # For backwards compatibility, but this only 
    # makes sense for Precomputed anyway
    return CloudVolumePrecomputed.create_new_info(*args, **kwargs)

  @classmethod
  def from_numpy(cls, 
    arr, 
    vol_path='file:///tmp/image/' + generate_random_string(),
    resolution=(4,4,40), voxel_offset=(0,0,0), 
    chunk_size=(128,128,64), layer_type=None, max_mip=0,
    encoding='raw', compress=None
  ):
    """
    Create a new dataset from a numpy array.

    max_mip: (int) the maximum mip level id in the info file. 
    Note that currently the numpy array can only sit in mip 0,
    the max_mip was only created in info file.
    the numpy array itself was not downsampled. 
    """
    if not layer_type:
      if arr.dtype in (np.bool, np.uint32, np.uint64, np.uint16):
        layer_type = 'segmentation'
      elif np.issubdtype(arr.dtype, np.integer) \
                        or np.issubdtype(arr.dtype, np.floating):
        layer_type = 'image'
      else:
        raise NotImplementedError

    if arr.ndim == 3:
      num_channels = 1
    elif arr.ndim == 4:
      num_channels = arr.shape[-1]
    else:
      raise NotImplementedError

    info = cls.create_new_info(
      num_channels, layer_type, arr.dtype.name,
      encoding, resolution,
      voxel_offset, arr.shape[:3],
      chunk_size=chunk_size, max_mip=max_mip
    )
    vol = CloudVolume(vol_path, info=info, bounded=True, compress=compress)
    # save the info file
    vol.commit_info()
    vol.provenance.processing.append({
      'method': 'from_numpy',
      'date': time.strftime('%Y-%m-%d %H:%M %Z')
    })
    vol.commit_provenance()
    # save the numpy array
    vol[:,:,:] = arr
    return vol



from functools import partial
import json
import os
import posixpath
import shutil

from . import scheduler

from .provenance import DataLayerProvenance
from cloudfiles import CloudFiles
from .storage.storage_interfaces import COMPRESSION_EXTENSIONS

from .lib import (
  Bbox, colorize, jsonify, mkdir, 
  toabs, Vec
)
from .paths import extract
from .secrets import CLOUD_VOLUME_DIR

def warn(text):
  print(colorize('yellow', text))

class CacheService(object):
  def __init__(
    self, cloudpath,
    enabled, config,
    meta=None, compress=None
  ):
    """
    enabled: bool or path string
    config: SharedConfiguration 
    meta: PrecomputedMetadata
    compress: None = linked to dataset setting, bool = Force
    """
    self._path = extract(cloudpath)
    self.path = self.default_path()

    self.config = config
    self._enabled = enabled 
    self.compress = compress 

    # b/c there's a semi-circular dependency
    # meta is usually set afterwards
    self.meta = meta 

    self.initialize()

  def initialize(self):
    if not self.enabled:
      return 

    if not os.path.exists(self.path):
      mkdir(self.path)

    if not os.access(self.path, os.R_OK|os.W_OK):
      raise IOError('Cache directory needs read/write permission: ' + self.path)

  @property
  def enabled(self):
    return self._enabled

  @enabled.setter
  def enabled(self, val):
    self._enabled = val 
    self.initialize()

  def default_path(self):
    basepath = self._path.basepath
    if basepath[0] == os.path.sep:
      basepath = basepath[1:]

    return toabs(os.path.join(CLOUD_VOLUME_DIR, 'cache', 
      self._path.protocol, basepath, self._path.layer
    ))
  
  def num_files(self, all_mips=False):
    def size(mip):
      path_at_mip = os.path.join(self.path, self.meta.key(mip))
      if not os.path.exists(path_at_mip):
        return 0
      return len(os.listdir(path_at_mip))

    if all_mips:
      sizes = [ 0 ] * (max(list(self.meta.available_mips)) + 1)
      for i in self.meta.available_mips:
        sizes[i] = size(i)
      return sizes
    else:
      return size(self.config.mip)

  def num_bytes(self, all_mips=False):
    def mip_size(mip):
      path_at_mip = os.path.join(self.path, self.meta.key(mip))
      if not os.path.exists(path_at_mip):
        return 0

      return sum( 
        ( os.path.getsize(os.path.join(path_at_mip, filename)) for filename in os.listdir(path_at_mip) ) 
      )

    if all_mips:
      sizes = [ 0 ] * (max(list(self.meta.available_mips)) + 1)
      for i in self.meta.available_mips:
        sizes[i] = mip_size(i)
      return sizes
    else:
      return mip_size(self.config.mip)

  def has(self, filename):
    return CloudFiles('file://' + self.path).exists(filename)
    
  def list(self, mip=None):
    mip = self.config.mip if mip is None else mip

    path = os.path.join(self.path, self.meta.key(mip))

    if not os.path.exists(path):
      return []

    return os.listdir(path)

  def list_skeletons(self):
    if self.meta.skeletons is None:
      return []

    path = os.path.join(self.path, self.meta.skeletons)
    if not os.path.exists(path):
      return []

    return os.listdir(path)

  def list_meshes(self):
    if self.meta.mesh is None:
      return []

    path = os.path.join(self.path, self.meta.mesh)
    if not os.path.exists(path):
      return []

    return os.listdir(path)

  def flush_info(self):
    path = os.path.join(self.path , 'info')
    if not os.path.exists(path):
      return
    os.remove(path)

  def flush_provenance(self):
    path = os.path.join(self.path , 'provenance')
    if not os.path.exists(path):
      return
    os.remove(path)

  def flush(self, preserve=None):
    """
    Delete the cache for this dataset. Optionally preserve
    a region. Helpful when working with overlaping volumes.

    Warning: the preserve option is not multi-process safe.
    You're liable to end up deleting the entire cache.

    Optional:
      preserve (Bbox: None): Preserve chunks located partially
        or entirely within this bounding box. 
    
    Return: void
    """
    if not os.path.exists(self.path):
      return

    if preserve is None:
      shutil.rmtree(self.path)
      return

    for mip in self.meta.available_mips:
      preserve_mip = self.meta.bbox_to_mip(preserve, 0, mip)
      mip_path = os.path.join(self.path, self.meta.key(mip))

      if not os.path.exists(mip_path):
        continue

      for filename in os.listdir(mip_path):
        bbox = Bbox.from_filename(filename)
        if not Bbox.intersects(preserve_mip, bbox):
          os.remove(os.path.join(mip_path, filename))

  # flush_cache_region seems like it could be tacked on
  # as a flag to delete, but there are reasons not
  # to do that. 
  # 1) reduces the risks of disasterous programming errors. 
  # 2) doesn't require chunk alignment
  # 3) processes potentially multiple mips at once

  def flush_region(self, region, mips=None):
    """
    Delete a cache region at one or more mip levels 
    bounded by a Bbox for this dataset. Bbox coordinates
    should be specified in mip 0 coordinates.

    Required:
      region (Bbox): Delete cached chunks located partially
        or entirely within this bounding box. 
    Optional:
      mip (int: None): Flush the cache from this mip. Region
        is in global coordinates.
    
    Return: void
    """
    if not os.path.exists(self.path):
      return
  
    cur_mip = self.config.mip

    region = Bbox.create(region, self.meta.bounds(cur_mip))
    mips = ( cur_mip, ) if mips == None else mips

    for mip in mips:
      mip_path = os.path.join(self.path, self.meta.key(mip))
      if not os.path.exists(mip_path):
        continue

      region_mip = self.meta.bbox_to_mip(region, mip=0, to_mip=mip)
      for filename in os.listdir(mip_path):
        bbox = Bbox.from_filename(filename)
        if not Bbox.intersects(region, bbox):
          os.remove(os.path.join(mip_path, filename))

  def check_info_validity(self):
    """
    ValueError if cache differs at all from source data layer with
    an excepton for volume_size which prints a warning.
    """
    cache_info = self.get_json('info')
    if not cache_info:
      return
    fresh_info = self.meta.fetch_info()

    mismatch_error = ValueError("""
      Data layer info file differs from cache. Please check whether this
      change invalidates your cache. 

      If VALID do one of:
      1) Manually delete the cache (see location below)
      2) Refresh your on-disk cache as follows:
        vol = CloudVolume(..., cache=False) # refreshes from source
        vol.cache = True
        vol.commit_info() # writes to disk
      If INVALID do one of: 
      1) Delete the cache manually (see cache location below) 
      2) Instantiate as follows: 
        vol = CloudVolume(..., cache=False) # refreshes info from source
        vol.flush_cache() # deletes cache
        vol.cache = True
        vol.commit_info() # writes info to disk

      CACHED: {cache}
      SOURCE: {source}
      CACHE LOCATION: {path}
      """.format(
      cache=cache_info, 
      source=fresh_info, 
      path=self.path
    ))

    try:
      fresh_sizes = [ scale['size'] for scale in fresh_info['scales'] ]
      cache_sizes = [ scale['size'] for scale in cache_info['scales'] ]
    except KeyError:
      raise mismatch_error

    for scale in fresh_info['scales']:
      del scale['size']

    for scale in cache_info['scales']:
      del scale['size']

    if fresh_info != cache_info:
      raise mismatch_error

    if fresh_sizes != cache_sizes:
      warn("WARNING: Data layer bounding box differs in cache.\nCACHED: {}\nSOURCE: {}\nCACHE LOCATION:{}".format(
      cache_sizes, fresh_sizes, self.path
      ))

  def check_provenance_validity(self):
    try:
      cached_prov = self.get_json('provenance')
    except json.decoder.JSONDecodeError:
      warn("Cached provenance file is not valid JSON.")
      return

    if not cached_prov:
      return

    cached_prov = self.meta._cast_provenance(cached_prov)
    fresh_prov = self.meta.fetch_provenance()
    if cached_prov != fresh_prov:
      warn("""
      WARNING: Cached provenance file does not match source.

      CACHED: {}
      SOURCE: {}
      """.format(cached_prov.serialize(), fresh_prov.serialize()))

  def get_json(self, filename):
    return CloudFiles('file://' + self.path).get_json(filename)

  def maybe_cache_info(self):
    if self.enabled:
      cf = CloudFiles('file://' + self.path)
      cf.put_json('info', self.meta.info)

  def maybe_cache_provenance(self):
    if self.enabled and self.meta.provenance:
      cf = CloudFiles('file://' + self.path)
      cf.put('provenance', self.meta.provenance.serialize().encode('utf8'))

  def upload_single(self, filename, content, *args, **kwargs):
    kwargs['progress'] = False
    return self.upload( [(filename, content)], *args, **kwargs )

  def upload(self, files, compress, cache_control, compress_level=None, content_type=None, progress=None):
    files = list(files)

    progress = progress if progress is not None else self.config.progress

    cf = CloudFiles(self.meta.cloudpath, progress=progress)
    cf.puts( 
      files, 
      compress=compress, 
      compression_level=compress_level, 
      cache_control=cache_control, 
      content_type=content_type 
    )

    if self.enabled:
      self.put(files, compress=compress)

  def download_json(self, path, compress=None):
    """
    Download a single path, but grab from 
    cache first if present and cache is enabled.

    Returns: content or None
    """
    res = self.download( [ path ], compress=compress, progress=False )
    res = res[path]
    if res is None:
      return None    
    return json.loads(res.decode('utf8'))

  def download_single(self, path, compress=None):
    files = self.download([ path ], compress=compress, progress=False)
    return files[path]

  def download_single_as(
    self, path, local_alias, 
    compress=None, start=None, end=None
  ):
    """
    Download a file or a byte range from a file 
    and save it locally as `local_alias`.
    """
    if self.enabled:
      locs = self.compute_data_locations([local_alias])
      if locs['local']:
        return self.get_single(local_alias)

    filedata = CloudFiles(self.meta.cloudpath)[path, start:end]

    if self.enabled:
      self.put([ (local_alias, filedata) ], compress=compress)

    return filedata

  def download_as(self, requests, compress=None, progress=None):
    """
    Works with byte ranges.

    requests: [{
      'path': ...,
      'local_alias': ...,
      'start': ...,
      'end': ...,
    }]
    """
    if len(requests) == 0:
      return {}

    progress = progress if progress is not None else self.config.progress

    aliases = [ req['local_alias'] for req in requests ]
    alias_tuples = { req['local_alias']: (req['path'], req['start'], req['end']) for req in requests }
    alias_to_path = { req['local_alias']: req['path'] for req in requests }
    path_to_alias = { v:k for k,v in alias_tuples.items() }

    if None in alias_to_path:
      del alias_to_path[None]
      del alias_tuples[None]

    locs = self.compute_data_locations(aliases)

    fragments = {}
    if self.enabled:
      fragments = self.get(locs['local'], progress=progress)
      keys = list(fragments.keys())
      for key in keys:
        return_key = (alias_to_path[key], alias_tuples[key][1], alias_tuples[key][2])
        fragments[return_key] = fragments[key]
        del fragments[key]
      for alias in locs['local']:
        del alias_tuples[alias]

    remote_path_tuples = list(alias_tuples.values())

    cf = CloudFiles(self.meta.cloudpath, progress=progress)
    remote_fragments = cf.get(
      ( { 'path': p[0], 'start': p[1], 'end': p[2] } for p in remote_path_tuples )
    )

    for frag in remote_fragments:
      if frag['error'] is not None:
        raise frag['error']

    remote_fragments = { 
      (res['path'], res['byte_range'][0], res['byte_range'][1]): res['content'] \
      for res in remote_fragments 
    }

    if self.enabled:
      self.put(
        [ 
          (path_to_alias[file_bytes_tuple], content) \
          for file_bytes_tuple, content in remote_fragments.items() \
          if content is not None 
        ],
        compress=compress,
        progress=progress
      )

    fragments.update(remote_fragments)
    return fragments

  def download(self, paths, compress=None, progress=None):
    """
    Download the provided paths, but grab them from cache first
    if they are present and the cache is enabled. 

    Returns: { filename: content, ... }
    """
    if len(paths) == 0:
      return {}

    progress = progress if progress is not None else self.config.progress

    locs = self.compute_data_locations(paths)
    locs['remote'] = [ str(x) for x in locs['remote'] ]

    fragments = {}
    if self.enabled:
      fragments = self.get(locs['local'], progress=progress)

    cf = CloudFiles(self.meta.cloudpath, progress=progress)
    remote_fragments = cf.get(locs['remote'])

    for frag in remote_fragments:
      if frag['error'] is not None:
        raise frag['error']

    remote_fragments = { 
      res['path']: res['content'] \
      for res in remote_fragments 
    }

    if self.enabled:
      self.put(
        [ 
          (filename, content) for filename, content in remote_fragments.items() \
          if content is not None 
        ],
        compress=compress,
        progress=progress
      )

    fragments.update(remote_fragments)
    return fragments

  def get_single(self, cloudpath, progress=None):
    res = self.get([ cloudpath ], progress=progress)
    return res[cloudpath]

  def get(self, cloudpaths, progress=None):
    progress = self.config.progress if progress is None else progress
    
    cf = CloudFiles('file://' + self.path, progress=progress)
    results = cf.get(list(cloudpaths))

    return { res['path']: res['content'] for res in results }

  def put_single(self, path, content, *args, **kwargs):
    kwargs['progress'] = False
    return self.put([ (path, content) ], *args, **kwargs)

  def put(self, files, progress=None, compress=None, compress_level=None):
    """files: [ (filename, content) ]"""
    if progress is None:
      progress = self.config.progress

    if compress is None:
      compress = self.compress

    if compress is None:
      compress = self.config.compress
    
    save_location = 'file://' + self.path
    progress = 'to Cache' if progress else None
    cf = CloudFiles(save_location, progress=progress)
    cf.puts(
      [ (name, content) for name, content in files ], 
      compress=compress, 
      compression_level=compress_level
    )

  def compute_data_locations(self, cloudpaths):
    if not self.enabled:
      return { 'local': [], 'remote': cloudpaths }

    pathmodule = posixpath if self.meta.path.protocol != 'file' else os.path

    def no_compression_ext(fnames):
      results = []
      for fname in fnames:
        (name, ext) = pathmodule.splitext(fname)
        if ext in COMPRESSION_EXTENSIONS:
          results.append(name)
        else:
          results.append(fname)
      return results

    list_dirs = set([ pathmodule.dirname(pth) for pth in cloudpaths ])
    filenames = []

    for list_dir in list_dirs:
      list_dir = os.path.join(self.path, list_dir)
      filenames += no_compression_ext(os.listdir(mkdir(list_dir)))

    basepathmap = { pathmodule.basename(path): pathmodule.dirname(path) for path in cloudpaths }

    # check which files are already cached, we only want to download ones not in cache
    requested = set([ pathmodule.basename(path) for path in cloudpaths ])
    already_have = requested.intersection(set(filenames))
    to_download = requested.difference(already_have)

    download_paths = [ pathmodule.join(basepathmap[fname], fname) for fname in to_download ]    
    already_have = [ os.path.join(basepathmap[fname], fname) for fname in already_have ]

    return { 'local': already_have, 'remote': download_paths }

  def __repr__(self):
    return "CacheService(enabled={}, compress={}, path='{}')".format(
      self.enabled, self.compress, self.path
    )



import pytest

import numpy as np

from cloudvolume.exceptions import DecompressionError
from cloudvolume.compression import compress, decompress

@pytest.mark.parametrize("compression_method", ("gzip", "br"))
def test_compression(compression_method):
  for N in range(100):
    flts = np.array(range(N), dtype=np.float32).reshape( (N,1,1,1) ).tostring()
    compressed = compress(flts, compression_method)
    assert compressed != flts
    decompressed = decompress(compressed, compression_method)
    assert decompressed == flts

  flts = np.array([], dtype=np.float32).tobytes()
  try:
    decompress(flts, compression_method)
    assert False
  except DecompressionError:
    pass

@pytest.mark.parametrize("compression_method", ("gzip", "br"))
def test_compress_level(compression_method):
  N = 10000
  x = np.array(range(N), dtype=np.float32).reshape( (N, 1, 1, 1) )
  content = np.ascontiguousarray(x, dtype=np.float32).tostring()

  compr_rate = []
  compress_levels = (1, 8)
  for compress_level in compress_levels:
    print(compress_level)
    compressed = compress(content, compression_method, compress_level=compress_level)

    assert compressed != content
    decompressed = decompress(compressed, compression_method)
    assert decompressed == content

    compr_rate.append(float(len(compressed)) / float(len(content)))

  # make sure we get better compression at highest level than lowest level
  assert compr_rate[-1] < compr_rate[0]


import six

from collections import defaultdict
import itertools
import json
import os
import posixpath
import re
import requests

import numpy as np
from tqdm import tqdm

from cloudfiles import CloudFiles

from ....lib import red, toiter, Bbox, Vec
from ....mesh import Mesh
from .... import paths

from ..sharding import GrapheneShardReader
from ...precomputed.sharding import ShardingSpecification

from .unsharded import GrapheneUnshardedMeshSource

class GrapheneShardedMeshSource(GrapheneUnshardedMeshSource):
  def __init__(self, mesh_meta, cache, config, readonly):
    super(GrapheneShardedMeshSource, self).__init__(mesh_meta, cache, config, readonly)

    self.readers = {}
    for level, sharding in self.meta.info['sharding'].items(): # { level: std sharding, ... }
      spec = ShardingSpecification.from_dict(sharding)
      self.readers[int(level)] = GrapheneShardReader(self.meta, self.cache, spec)

  def initial_path(self, level):
    return self.meta.join(self.meta.mesh_path, 'initial', str(level))

  def dynamic_path(self):
    return self.meta.join(self.meta.mesh_path, 'dynamic')

  # 1. determine if the segid is before or after the shard time point
  # 2. assuming it is sharded, fetch the draco encoded file from the
  #    correct level

  def dynamic_exists(self, labels, progress=None):
    """
    Checks for dynamic mesh existence.
  
    Returns: { label: path or None, ... }
    """
    labels = toiter(labels)

    checks = [ self.compute_filename(label) for label in labels ]
    
    cloudpath = self.meta.join(self.meta.meta.cloudpath, self.meta.mesh_path, 'dynamic') 
    progress = progress if progress is not None else self.config.progress

    results = CloudFiles(cloudpath, progress=progress, green=self.config.green).exists(checks)

    output = {}
    for filepath, exists in results.items():
      label = int(os.path.basename(filepath)[:-2]) # strip :0
      output[label] = filepath if exists else None

    return output

  def initial_exists(self, labels, return_byte_range=False, progress=None):
    """
    Checks for initial mesh existence.
  
    Returns: 
      If return_byte_range:
        { label: [ path, byte offset, byte size ] or None, ... }
      Else:
        { label: path or None, ... }
    """
    labels = toiter(labels)
    progress = progress if progress is not None else self.config.progress

    layers = defaultdict(list)
    for label in labels:
      if label == 0:
        continue
      layer = self.meta.meta.decode_layer_id(label)
      layers[layer].append(label)

    all_results = {}
    for layer, layer_labels in layers.items():
      path = self.initial_path(layer)
      results = self.readers[int(layer)].exists(
        layer_labels, 
        path=path, 
        return_byte_range=return_byte_range, 
        progress=progress
      )
      all_results.update(results)

    return all_results

  def exists(self, labels, progress=None):
    """
    Checks dynamic then initial meshes for existence.

    Returns: { label: path or None, ... }
    """
    labels = toiter(labels)
    labels = set(labels)

    dynamic_labels = self.dynamic_exists(labels, progress)
    remainder_labels = set([ label for label, path in dynamic_labels.items() if path ])

    initial_labels = self.initial_exists(remainder_labels, progress=progress, return_byte_range=False)

    dynamic_labels.update(initial_labels)
    return dynamic_labels

  def parse_manifest_filenames(self, filenames):
    lists = defaultdict(list)
    initial_regexp = re.compile(r'~(\d+)/([\d\-]+\.shard):(\d+):(\d+)')

    for filename in filenames:
      if not filename:
        continue

      # eg. ~2/344239114-0.shard:224659:442 
      # tilde means initial, missing tilde means dynamic
      initial = filename[0] == '~'

      if initial:
        (layer_id, parsed_filename, byte_start, size) = re.search(
          initial_regexp, filename
        ).groups()
        lists['initial'].append((layer_id, parsed_filename, int(byte_start), int(size)))
      else:
        lists['dynamic'].append(filename)
        
    return lists

  def get_meshes_via_manifest_byte_offsets(self, seg_id, bounding_box):
    """    
    The manifest for sharded is a bit strange in that exists(..., return_byte_offset=True)
    is being called on the server side. To avoid duplicative delay by recomputing the offset
    locations, the manifest breaks encapsulation by returning the shard filename and byte
    offsets. This breaks enapsulation of the shard fetching logic rather severely but 
    it is probably worth it.
    """
    level = self.meta.meta.decode_layer_id(seg_id)
    dynamic_cloudpath = self.meta.join(self.meta.meta.cloudpath, self.dynamic_path())

    filenames = self.get_fragment_filenames(seg_id, level=level, bbox=bounding_box)
    lists = self.parse_manifest_filenames(filenames)

    files = []
    if lists['dynamic']:
      files = CloudFiles(dynamic_cloudpath, green=self.config.green).get(lists['dynamic'])

    meshes = [ 
      f['content'] for f in files 
    ]

    fetches = []
    for layer_id, filename, byte_start, size in lists['initial']:
      fetches.append({
        'path': self.meta.join(layer_id, filename),
        'start': byte_start,
        'end': byte_start + size,
      })

    cloudpath = self.meta.join(self.meta.meta.cloudpath, self.meta.mesh_path, 'initial')
    raw_binaries = []
    
    initial_meshes = CloudFiles(cloudpath, green=self.config.green).get(fetches)
    meshes += initial_meshes

    return [ Mesh.from_draco(mesh['content']) for mesh in meshes ]

  def get_meshes_via_manifest_labels(self, seg_id, bounding_box):
    level = self.meta.meta.decode_layer_id(seg_id)
    labels = self.get_fragment_labels(seg_id, level=level, bbox=bounding_box)
    meshes = self.get_meshes_on_bypass(labels)
    return list(meshes.values())

  def get_meshes_via_manifest(self, seg_id, bounding_box, use_byte_offsets):
    if use_byte_offsets:
      return self.get_meshes_via_manifest_byte_offsets(seg_id, bounding_box)
    return self.get_meshes_via_manifest_labels(seg_id, bounding_box)

  def get_meshes_on_bypass(self, segids):
    """
    Attempt to fetch a mesh directly from storage without going through
    the chunk graph server. This capability should only be used in special
    circumstances.
    """
    segids = toiter(segids)
    
    dynamic_cloudpath = self.meta.join(self.meta.meta.cloudpath, self.dynamic_path())
    filenames = [ self.compute_filename(segid) for segid in segids ]

    cf = CloudFiles(dynamic_cloudpath, progress=self.config.progress, green=self.config.green)
    raw_binaries = cf.get(filenames)

    # extract the label ID from the mesh manifest.
    # e.g. 387463568301300850:0:24576-25088_17920-18432_2048-3072
    label_regexp = re.compile(r'(\d+):\d:[\d_-]+$')

    output = {}
    remaining = []
    for res in raw_binaries:
      if res['error']:
        raise res['error']

      (label,) = re.search(label_regexp, res['filename']).groups()
      label = int(label)

      if res['content'] is None:
        remaining.append(label)
      else:
        output[label] = res['content']

    layers = defaultdict(list)
    for segid in remaining:
      layer_id = self.meta.meta.decode_layer_id(segid)
      layers[layer_id].append(segid)

    for layer_id, labels in layers.items():
      subdirectory = self.meta.join(self.meta.mesh_path, 'initial', str(layer_id))
      initial_output = self.readers[layer_id].get_data(labels, path=subdirectory, progress=self.config.progress)
      for label, raw_binary in initial_output.items():
        if raw_binary is None:
          raise IndexError('No mesh found for segment {}'.format(label))
      output.update(initial_output)

    return { label: Mesh.from_draco(raw_binary) for label, raw_binary in output.items() }

  def download_segid(self, seg_id, bounding_box, bypass=False, use_byte_offsets=True):    
    """See GrapheneUnshardedMeshSource.get for the user facing function."""
    level = self.meta.meta.decode_layer_id(seg_id)
    if level not in self.readers:
      raise KeyError("There is no shard configuration in the mesh info file for level {}.".format(level))

    if bypass:
      mesh = self.get_meshes_on_bypass(seg_id)[seg_id]
    else:
      meshes = self.get_meshes_via_manifest(seg_id, bounding_box, use_byte_offsets=use_byte_offsets)
      mesh = Mesh.concatenate(*meshes)

    if mesh is None:
      raise IndexError('No mesh found for segment {}'.format(seg_id))

    mesh.segid = seg_id
    return mesh, True



from collections import defaultdict
import json
import os 

import numpy as np

from cloudfiles import CloudFiles

from ...exceptions import SpatialIndexGapError
from ... import paths
from ...lib import Bbox, Vec, xyzrange, min2, toiter

class SpatialIndex(object):
  """
  Implements the client side reader of the 
  spatial index. During data generation, the
  labels in a given task are enumerated and 
  assigned their bounding box as JSON:

  {
    SEGID: [ x,y,z, x,y,z ],
    ...
  }

  The filename is the physical bounding box of the
  task dot spatial.

  e.g. "0-1024_0-1024_0-500.spatial" where the bbox 
  units are nanometers.

  The info file of the data type can then be augmented
  with:

  {
    "spatial_index": { "chunk_size": [ sx, sy, sz ] }
  }

  Where sx, sy, and sz are given in physical dimensions.
  """
  def __init__(self, cloudpath, bounds, chunk_size, progress=False):
    self.cloudpath = cloudpath
    self.path = paths.extract(cloudpath)
    self.bounds = Bbox.create(bounds)
    self.chunk_size = Vec(*chunk_size)
    self.progress = progress

  def join(self, *paths):
    if self.path.protocol == 'file':
      return os.path.join(*paths)
    else:
      return posixpath.join(*paths)    

  def fetch_index_files(self, index_files):
    results = CloudFiles(self.cloudpath, progress=self.progress).get(index_files)

    for res in results:
      if res['error'] is not None:
        raise SpatialIndexGapError(res['error'])

    return { res['filename']: res['content'] for res in results }

  def index_file_paths_for_bbox(self, bbox):
    bbox = bbox.expand_to_chunk_size(self.chunk_size, offset=self.bounds.minpt)

    if bbox.subvoxel():
      return []

    index_files = []
    for pt in xyzrange(bbox.minpt, bbox.maxpt, self.chunk_size):
      search = Bbox( pt, min2(pt + self.chunk_size, self.bounds.maxpt) )
      index_files.append(search.to_filename() + '.spatial')
    
    return index_files

  def get_bbox(self, label):
    """
    Given a label, compute an enclosing bounding box for it.

    Returns: Bbox in physical coordinates
    """
    index_files = self.index_file_paths_for_bbox(self.bounds)
    index_files = self.fetch_index_files(index_files)
    locations = defaultdict(list)
    
    label = str(label)
    bbox = None
    for filename, content in index_files.items():
      if content is None:
        if allow_missing:
          continue
        else:
          raise SpatialIndexGapError(filename + " was not found.")

      segid_bbox_dict = json.loads(content)
      filename = os.path.basename(filename)

      if label not in segid_bbox_dict: 
        continue 

      current_bbox = Bbox.from_list(segid_bbox_dict[label])

      if bbox is None:
        bbox = current_bbox
      else:
        bbox = Bbox.expand(bbox, current_bbox)

    return bbox

  def file_locations_per_label(self, labels=None, allow_missing=False):
    """
    Queries entire dataset to find which spatial index files the 
    given labels are located in. Can be expensive. If labels is not 
    specified, all labels are fetched.

    Returns: { filename: [ labels... ], ... }
    """
    if labels is not None:
      labels = set(toiter(labels))
      
    index_files = self.index_file_paths_for_bbox(self.bounds)
    index_files = self.fetch_index_files(index_files)
    locations = defaultdict(list)
    for filename, content in index_files.items():
      if content is None:
        if allow_missing:
          continue
        else:
          raise SpatialIndexGapError(filename + " was not found.")

      segid_bbox_dict = json.loads(content)
      filename = os.path.basename(filename)

      if labels is None:
        for label in segid_bbox_dict.keys():
          locations[int(label)].append(filename)
      elif len(labels) > len(segid_bbox_dict):
        for label in segid_bbox_dict.keys():
          if int(label) in labels:
            locations[int(label)].append(filename)
      else:
        for label in labels:
          if str(label) in segid_bbox_dict:
            locations[int(label)].append(filename)

    return locations

  def query(self, bbox, allow_missing=False):
    """
    For the specified bounding box (or equivalent representation),
    list all segment ids enclosed within it.

    If allow_missing is set, then don't raise an error if an index
    file is missing.

    Returns: set(labels)
    """
    bbox = Bbox.create(bbox, context=self.bounds, autocrop=True)
    original_bbox = bbox.clone()
    bbox = bbox.expand_to_chunk_size(self.chunk_size, offset=self.bounds.minpt)

    if bbox.subvoxel():
      return []

    index_files = self.index_file_paths_for_bbox(bbox)
    results = self.fetch_index_files(index_files)

    labels = set()
    for filename, content in results.items():
      if content is None:
        if allow_missing:
          continue
        else:
          raise SpatialIndexGapError(filename + " was not found.")

      res = json.loads(content)

      # The bbox test saps performance a lot
      # but we can skip it if we know 100% that
      # the labels are going to be inside. This
      # optimization is important for querying 
      # entire datasets, which is contemplated
      # for shard generation.
      if bbox.contains_bbox(self.bounds):
        for label, label_bbx in res.items():
          labels.add(int(label))
      else:
        for label, label_bbx in res.items():
          label = int(label)
          label_bbx = Bbox.from_list(label_bbx)

          if Bbox.intersects(label_bbx, original_bbox):
            labels.add(label)

    return labels

class CachedSpatialIndex(SpatialIndex):
  def __init__(self, cache, cloudpath, bounds, chunk_size, progress=None):
    self.cache = cache
    self.subdir = os.path.relpath(cloudpath, cache.meta.cloudpath)

    super(CachedSpatialIndex, self).__init__(
      cloudpath, bounds, chunk_size, progress
    )

  def fetch_index_files(self, index_files):
    index_files = [ self.cache.meta.join(self.subdir, fname) for fname in index_files ]
    return self.cache.download(index_files, progress=self.progress)



import os
import setuptools
import sys

def read(fname):
  with open(os.path.join(os.path.dirname(__file__), fname), 'rt') as f:
    return f.read()

def requirements():
  with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), 'rt') as f:
    return f.readlines()

setuptools.setup(
  name="cloud-volume",
  version="1.19.0",
  setup_requires=[
    'numpy<1.17; python_version<"3.5"',
    'numpy; python_version>="3.5"',
  ],
  install_requires=requirements(),
  # Environment Marker Examples:
  # https://www.python.org/dev/peps/pep-0496/
  extras_require={
    "boss": [
      "intern>=0.9.11",
      "blosc==1.8.3",
    ],
    ':sys_platform!="win32"': [
      "posix_ipc==1.0.4",
      "psutil==5.4.3",
    ],
    ':sys_platform!="win32" and python_version>="3.0"': [
      "DracoPy",
    ],
    "mesh_viewer": [ 'vtk' ],
    "skeleton_viewer": [ 'matplotlib' ],
    "all_viewers": [ 'vtk', 'matplotlib' ],
    "dask": [ 'dask[array]' ],
    "test": [ "pytest", "pytest-cov", "codecov", "requests_mock", "scipy"]
  },
  author="William Silversmith, Nico Kemnitz, Ignacio Tartavull, and others",
  author_email="ws9@princeton.edu",
  packages=setuptools.find_packages(),
  package_data={
    'cloudvolume': [
      './microviewer/*',
      'LICENSE',
    ],
  },
  description="A serverless client for reading and writing Neuroglancer Precomputed volumes both locally and on cloud services.",
  long_description=read('README.md'),
  long_description_content_type="text/markdown",
  license = "BSD 3-Clause",
  keywords = "neuroglancer volumetric-data numpy connectomics microscopy image-processing biomedical-image-processing s3 gcs mesh meshes skeleton skeletons",
  url = "https://github.com/seung-lab/cloud-volume/",
  classifiers=[
    "Intended Audience :: Developers",
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: BSD License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
    "Intended Audience :: Science/Research",
    "Operating System :: POSIX",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows :: Windows 10",
    "Topic :: Utilities",
  ],
)



import re
import subprocess
import sys

def get_latest_version():
	tags = str(subprocess.check_output('git tag', shell=True))
	tags = re.sub("^b'", '', tags)
	tags = re.sub("'$", '', tags)
	tags = re.split(r'\\n', tags)
	tags = [ tuple(map(int, re.split('\.', tag))) for tag in tags if tag ]
	tags = sorted(tags, reverse=True)
	version = [ str(_) for _ in tags[0] ]
	return '.'.join(version)

def materialize_version(version):
	# version = get_version()
	with open('cloudvolume/__init__.py', 'rt') as f:
		code = f.read()

	code = re.sub("__version__ = '.*?'", "__version__ = '{}'".format(version), code)
	with open('cloudvolume/__init__.py', 'wt') as f:
		f.write(code)

	with open('setup.py', 'rt') as f:
		code = f.read()

	code = re.sub(r'version\s*=\s*".*?",', 'version="{}",'.format(version), code)
	with open('setup.py', 'wt') as f:
		f.write(code)

version = sys.argv[1]
message = sys.argv[2]

materialize_version(version)
print("Updated __init__.py with version " + version)

assert len(version) < 9
subprocess.check_output("git reset && git add cloudvolume/__init__.py setup.py && git commit -m 'Version {}'".format(version), shell=True)

try:
	subprocess.check_output(['git', 'tag', '-a',  version, '-m', message])
	print('Created tag ' + version)
except:
	pass





class InfoError(Exception):
  pass

class InfoUnavailableError(InfoError):
  """CloudVolume was unable to access this layer's info file."""
  pass

class RedirectError(InfoError):
  pass

class TooManyRedirects(RedirectError):
  """The chain of redirects became unconvincingly long."""
  pass

class CyclicRedirect(RedirectError):
  """Unable to resolve redirects due to cycle."""
  pass

class ScaleUnavailableError(IndexError):
  """The info file is not configured to support this scale / mip level."""
  pass

class ReadOnlyException(Exception):
  """Attempted to write to a readonly data source."""
  pass

class AlignmentError(ValueError):
  """Signals that an operation requiring chunk alignment was not aligned."""
  pass 

class EmptyVolumeException(Exception):
  """Raised upon finding a missing chunk."""
  pass

class EmptyFileException(Exception):
  """File was zero bytes."""
  pass

class EmptyRequestException(ValueError):
  """
  Requesting uploading or downloading 
  a bounding box of less than one cubic voxel
  is impossible.
  """
  pass

class DecodingError(Exception):
  """Generic decoding error. Applies to content aware and unaware codecs."""
  pass

class EncodingError(Exception):
  """Generic decoding error. Applies to content aware and unaware codecs."""
  pass

class OutOfBoundsError(ValueError):
  """
  Raised upon trying to obtain or assign to a bbox of a volume outside
  of the volume's bounds
  """

class UnsupportedCompressionType(ValueError):
  """
  Raised when attempting to use a compression type which is unsupported
  by the storage interface.
  """

# Inheritance below done for backwards compatibility reasons.

class DecompressionError(DecodingError):
  """
  Decompression failed. This exception is used for codecs 
  that are naieve to data contents like gzip, lzma, etc. as opposed
  to codecs that are aware of array shape like fpzip or compressed_segmentation.
  """
  pass

class CompressionError(EncodingError):
  """
  Compression failed. This exception is used for codecs 
  that are naieve to data contents like gzip, lzma, etc. as opposed
  to codecs that are aware of array shape like fpzip or compressed_segmentation.
  """
  pass

class MeshDecodeError(ValueError):
  """Unable to decode a mesh object."""
  pass

class SkeletonUnassignedEdgeError(Exception):
  """This skeleton has an edge to a vertex that doesn't exist."""
  pass

class SkeletonDecodeError(Exception):
  """Unable to decode a binary skeleton into a Python object."""
  pass

class SkeletonEncodeError(Exception):
  """Unable to encode a PrecomputedSkeleton into a binary object."""
  pass

class SkeletonTransformError(Exception):
  """Unable to apply a spatial transfrom to the current coordinate system."""
  pass

class SkeletonAttributeMixingError(Exception):
  """
  These skeletons have different vertex attributes 
  and cannot be recombined without manual intervention.
  """
  pass

class UnsupportedFormatError(Exception):
  """Unable to interpret the format of this URI. e.g. precomputed://"""
  pass

class UnsupportedProtocolError(ValueError):
  """Unknown protocol extension."""
  pass

class UnsupportedGrapheneAPIVersionError(Exception):
  """This dataset does not support the specified api version."""
  pass

class SpecViolation(Exception):
  """The values held by this object violate its written specification."""
  pass

class SpatialIndexGapError(Exception):
  """Part of the spatial index was not found. A complete result set cannot be fetched."""
  pass

class WriteLockAcquisitionError(Exception):
  """Unable to obtain a lock on this data layer element."""
  pass

class WriteLockReleaseError(Exception):
  """Unable to release a lock on this data layer element."""
  pass